import numpy as np
from copy import deepcopy
from sys import version_info

from occamypy import Vector


class VectorNumpy(Vector):
    """In-core python vector class"""
    
    def __init__(self, in_content):
        """
        VectorNumpy constructor: arr = np.array
        This class stores array with C memory order (i.e., row-wise sorting)
        """
        if isinstance(in_content, str):  # Header file name passed to constructor
            self.arr = np.load(in_content, allow_pickle=True)
        elif isinstance(in_content, np.ndarray):  # Numpy array passed to constructor
            if np.isfortran(in_content):
                raise TypeError('Input array not a C contiguous array!')
            self.arr = np.array(in_content, copy=False)
            self.ax_info = None
        elif isinstance(in_content, tuple):  # Tuple size passed to constructor
            # self.arr = np.zeros(tuple(reversed(in_vec)))
            self.arr = np.empty(in_content)
            self.ax_info = None
        else:  # Not supported type
            raise ValueError("ERROR! Input variable not currently supported!")
        
        super(VectorNumpy, self).__init__()
        # Number of elements per axis (tuple). Checking also the memory order
        self.shape = self.arr.shape  # If fortran the first axis is the "fastest"
        self.ndim = self.arr.ndim  # Number of axes integer
        self.size = self.arr.size  # Total number of elements
    
    def getNdArray(self):
        """Function to return Ndarray of the vector"""
        return self.arr
    
    def norm(self, N=2):
        """Function to compute vector N-norm using Numpy"""
        return np.linalg.norm(self.getNdArray().ravel(), ord=N)
    
    def zero(self):
        """Function to zero out a vector"""
        self.getNdArray().fill(0)
        return self
    
    def max(self):
        """Function to obtain maximum value in the vector"""
        return self.getNdArray().max()
    
    def min(self):
        """Function to obtain minimum value in the vector"""
        return self.getNdArray().min()
    
    def set(self, val):
        """Function to set all values in the vector"""
        self.getNdArray().fill(val)
        return self
    
    def scale(self, sc):
        """Function to scale a vector"""
        self.getNdArray()[:] *= sc
        return self
    
    def addbias(self, bias):
        self.getNdArray()[:] += bias
        return self
    
    def rand(self, snr=1.):
        """Fill vector with random number (~U[1,-1]) with a given SNR"""
        rms = np.sqrt(np.mean(np.square(self.getNdArray())))
        amp_noise = 1.0
        if rms != 0.:
            amp_noise = np.sqrt(3. / snr) * rms  # sqrt(3*Power_signal/SNR)
        self.getNdArray()[:] = amp_noise * (2. * np.random.random(self.getNdArray().shape) - 1.)
        return self
    
    def clone(self):
        """Function to clone (deep copy) a vector from a vector or a Space"""
        vec_clone = deepcopy(self)  # Deep clone of vector
        # Checking if a vector space was provided
        if vec_clone.getNdArray().size == 0:  # this is the shape of np.ndarray!
            vec_clone.arr = np.zeros(vec_clone.shape, dtype=self.getNdArray().dtype)
        return vec_clone
    
    def cloneSpace(self):
        """Function to clone vector space only (vector without actual vector array by using empty array of size 0)"""
        vec_space = VectorNumpy(np.empty(0, dtype=self.getNdArray().dtype))
        # Cloning space of input vector
        vec_space.ndim = self.ndim
        vec_space.shape = self.shape
        vec_space.size = self.size
        return vec_space
    
    def checkSame(self, other):
        """Function to check dimensionality of vectors"""
        return self.shape == other.shape
    
    def abs(self):
        self.getNdArray()[:] = np.abs(self.getNdArray())
        return self
    
    def sign(self):
        self.getNdArray()[:] = np.sign(self.getNdArray())
        return self
    
    def reciprocal(self):
        self.getNdArray()[:] = 1. / self.getNdArray()
        return self
    
    def maximum(self, other):
        if np.isscalar(other):
            self.getNdArray()[:] = np.maximum(self.getNdArray(), other)
            return self
        elif isinstance(other, VectorNumpy):
            if not self.checkSame(other):
                raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
            self.getNdArray()[:] = np.maximum(self.getNdArray(), other.getNdArray())
            return self
        else:
            raise TypeError("Provided input has to be either a scalar or a %s!" % self.whoami)
    
    def conj(self):
        self.getNdArray()[:] = np.conjugate(self.getNdArray())
        return self
    
    def transpose(self):
        other = VectorNumpy(tuple(reversed(self.shape)))
        other[:] = self.getNdArray().T
        return other
    
    def pow(self, power):
        """Compute element-wise power of the vector"""
        self.getNdArray()[:] = self.getNdArray() ** power
        return self
    
    def real(self):
        """Return the real part of the vector"""
        self.getNdArray()[:] = self.getNdArray().real
        return self
    
    def imag(self, ):
        """Return the imaginary part of the vector"""
        self.getNdArray()[:] = self.getNdArray().imag
        return self
    
    def copy(self, other):
        """Function to copy vector from input vector"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorNumpy):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
        # Element-wise copy of the input array
        self.getNdArray()[:] = other.getNdArray()
        return self
    
    def scaleAdd(self, other, sc1=1.0, sc2=1.0):
        """Function to scale a vector"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorNumpy):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
        # Performing scaling and addition
        self.getNdArray()[:] = sc1 * self.getNdArray() + sc2 * other.getNdArray()
        return self
    
    def dot(self, other):
        """Function to compute dot product between two vectors"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorNumpy):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking size (must have same number of elements)
        if self.size != other.size:
            raise ValueError("Vector size mismatching: self = %d; other = %d" % (self.size, other.size))
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
        return np.vdot(self.getNdArray().ravel(), other.getNdArray().ravel())
    
    def multiply(self, other):
        """Function to multiply element-wise two vectors"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorNumpy):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking size (must have same number of elements)
        if self.size != other.size:
            raise ValueError("Vector size mismatching: self = %s; other = %s" % (self.size, other.size))
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
        # Performing element-wise multiplication
        self.getNdArray()[:] = np.multiply(self.getNdArray(), other.getNdArray())
        return self
    
    def isDifferent(self, other):
        """Function to check if two vectors are identical using built-in hash function"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorNumpy):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Using Hash table for python2 and numpy built-in function array_equal otherwise
        if version_info[0] == 2:
            # First make both array buffers read-only
            self.arr.flags.writeable = False
            other.arr.flags.writeable = False
            chcksum1 = hash(self.getNdArray().data)
            chcksum2 = hash(other.getNdArray().data)
            # Remake array buffers writable
            self.arr.flags.writeable = True
            other.arr.flags.writeable = True
            isDiff = (chcksum1 != chcksum2)
        else:
            isDiff = (not np.array_equal(self.getNdArray(), other.getNdArray()))
        return isDiff
    
    def clipVector(self, low, high):
        """Function to bound vector values based on input vectors low and high"""
        if not isinstance(low, VectorNumpy):
            raise TypeError("Provided input low vector not a %s!" % self.whoami)
        if not isinstance(high, VectorNumpy):
            raise TypeError("Provided input high vector not a %s!" % self.whoami)
        self.getNdArray()[:] = np.minimum(np.maximum(low.getNdArray(), self.getNdArray()), high.getNdArray())
        return self



import numpy as np

from occamypy.operator import Operator, Vstack
from occamypy.operator.base import _sumOperator, _prodOperator


def dummy_set_background(dummy_arg):
    """
    Dummy function to use Non-linear operator class for Linear ones (it takes one argument and does nothing)
    """
    return


class NonlinearOperator(Operator):
    """
    Non-linear operator class
    """
    
    def __init__(self, nl_op, lin_op=None, set_background_func=dummy_set_background):
        """
           Constructor for non-linear operator class:
           nl_op                = [no default] - operator class;
                                Non-linear operator class where only the forward is overwritten
           lin_op               = [no default] - operator class;
                                Linear Jacobian operator class where only the forward is
                                overwritten (if not necessary, use pyOperator.ZeroOp)
           set_background_func  = [dummy_set_background] - function pointer;
                                Function to set the model vector on which the
                                Jacobian operator is evaluated
        """
        # Setting non-linear and linearized operators
        self.nl_op = nl_op
        self.lin_op = lin_op if lin_op is not None else nl_op
        self.set_background = set_background_func
        # Checking if domain of the operators is the same
        if not self.nl_op.domain.checkSame(self.lin_op.domain):
            raise ValueError("ERROR! The two provided operators have different domains")
        if not self.nl_op.range.checkSame(self.lin_op.range):
            raise ValueError("ERROR! The two provided operators have different ranges")
        super(NonlinearOperator, self).__init__(self.nl_op.domain, self.nl_op.range)
    
    def dotTest(self, **kwargs):
        """
        Raising an exception, dot-product tests must be performed directly onto linear operator.
        """
        raise NotImplementedError("Perform dot-product tests directly on the linear operator.")
    
    def linTest(self, background, pert=None, alpha=np.logspace(-6, 0, 100), plot=False):
        """
        Linearization tests function. It plots the model-perturbation norm vs linearization error norm
        :param background: vector class - Background model used during the linearization tests
        :param pert: vector class - Model-perturbation vector to be used during linearization tests [None]
                     if not provided a random perturbation is employed
        :param alpha: array - array of scalars to scale the pert vector during the tests [np.logspace(-6,0,100)]
        :param plot: boolean - whether to plot the linearization error vs perturbation scale or not
        :return:
        :param scale
        :param lin_err array - array containing linearization error for each scale value
        """
        # Creating model perturbation if not provided
        if pert is None:
            pert = self.getDomain().clone().rand()
        # List containing linearization error and perturbation scale/norm
        lin_err = []
        # temporary vectors
        m0 = background.clone()
        m = background.clone()
        d0 = self.nl_op.getRange().clone()
        d1 = self.nl_op.getRange().clone()
        dlin = self.nl_op.getRange().clone()
        # computing f(m0) = d0
        self.nl_op.forward(False, m0, d0)
        # setting m0 for the Jacobian matrix
        self.set_background(m0)
        # computing F(m0)dm = dlin
        self.lin_op.forward(False, pert, dlin)
        for sc in alpha:
            # print(sc)
            # computing f(m0+dm) = d1
            m.copy(m0)
            m.scaleAdd(pert, 1.0, sc)
            self.nl_op.forward(False, m, d1)
            # computing f(m0+dm) - f(m0)
            d1.scaleAdd(d0, 1.0, -1.0)
            # computing f(m0+dm) - f(m0) - F(m0)dm (i.e., linearization error)
            d1.scaleAdd(dlin, 1.0, -sc)
            lin_err.append(d1.norm())
        lin_err = np.array(lin_err)
        if plot:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(figsize=(6, 3))
            plt.plot(alpha, lin_err, 'r')
            ax.autoscale(enable=True, axis='y', tight=True)
            ax.autoscale(enable=True, axis='x', tight=True)
            plt.xlabel(r"$\alpha$")
            plt.ylabel(r"$|f(m_0+\alpha dm) - f(m_0) - \alpha F(m_0)dm|_2$")
            plt.title('Linearization error')
            plt.show()
        return alpha, lin_err
    
    # unary operators
    def __add__(self, other):  # self + other
        if isinstance(other, NonlinearOperator):
            return _sumNlOperator(self, other)
        else:
            raise TypeError('Argument must be an Operator')


class NonlinearComb(NonlinearOperator):
    """
    Combination of non-linear opeartors: f(g(m))
    """
    
    def __init__(self, f, g):
        """
        Constructor for non-linear operator class
        """
        # Checking if non-linear operators were provided
        if not (isinstance(f, NonlinearOperator) and isinstance(g, NonlinearOperator)):
            raise TypeError("Provided operators has to be NonLinearOperator")
        # Defining f(g(m))
        self.nl_op = _prodOperator(f.nl_op, g.nl_op)
        # Defining F(g(m0))G(m0)
        self.lin_op = _prodOperator(f.lin_op, g.lin_op)
        # Defining internal set_background functions
        self.set_background_f = f.set_background
        self.set_background_g = g.set_background
        # Defining non_linear operator g(m) for Jacobian definition
        self.g_nl_op = g.nl_op
        self.g_range_tmp = g.nl_op.range.clone()
        super(NonlinearComb, self).__init__(self.nl_op, self.lin_op, self.set_background)
    
    def set_background(self, model):
        """
        Set background function for the chain of Jacobian matrices
        """
        # Setting G(m0)
        self.set_background_g(model)
        # Setting F(g(m0))
        self.g_nl_op.forward(False, model, self.g_range_tmp)
        self.set_background_f(self.g_range_tmp)


# Necessary for backward compatibility
def CombNonlinearOp(g, f):
    """Combination of non-linear operators: f(g(m))"""
    return NonlinearComb(f, g)


class _sumNlOperator(NonlinearOperator):
    """
    Sum of two non-linear operators
        h = g + f
    """
    
    def __init__(self, g, f):
        """Sum operator constructor"""
        if not isinstance(g, NonlinearOperator) or not isinstance(f, NonlinearOperator):
            raise TypeError('Both operands have to be a NonLinearOperator')
        if not f.range.checkSame(g.range) or not f.domain.checkSame(g.domain):
            raise ValueError('Cannot add operators: shape mismatch')
        
        self.args = (g, f)
        # Defining f(m) + g(m)
        self.nl_op = _sumOperator(f.nl_op, g.nl_op)
        # Defining F(m0) and G(m0)
        self.lin_op = _sumOperator(f.lin_op, g.lin_op)
        # Defining internal set_background functions
        self.set_background_f = f.set_background
        self.set_background_g = g.set_background
        # Defining non_linear operator g(m) for Jacobian definition
        self.g_nl_op = g.nl_op
        self.g_range_tmp = g.nl_op.range.clone()
        super(_sumNlOperator, self).__init__(self.nl_op, self.lin_op, self.set_background)
    
    def __str__(self):
        return self.args[0].__str__()[:3] + "+" + self.args[1].__str__()[:4]
    
    def set_background(self, model):
        """
        Set background function for the sum of Jacobian matrices
        """
        # Setting G(m0)
        self.set_background_g(model)
        # Setting F(m0)
        self.set_background_f(model)


def NonlinearSum(f, g):
    """Combination of non-linear operators: f(m) + g(m)"""
    return _sumNlOperator(f, g)


class NonlinearVstack(NonlinearOperator):
    """
    Stack of operators class
            | d1 |   | f(m) |
     h(m) = |    | = |      |
            | d2 |   | g(m) |
    """
    
    def __init__(self, nl_op1, nl_op2):
        """Constructor for the stacked operator"""
        # Checking if domain of the operators is the same
        if not (isinstance(nl_op1, NonlinearOperator) and isinstance(nl_op2, NonlinearOperator)):
            raise TypeError("Provided operators must be NonLinearOperator instances")
        self.nl_op1 = nl_op1  # f(m)
        self.nl_op2 = nl_op2  # g(m)
        # Defining f(g(m))
        self.nl_op = Vstack(nl_op1.nl_op, nl_op2.nl_op)
        # Defining F(g(m0))G(m0)
        self.lin_op = Vstack(nl_op1.lin_op, nl_op2.lin_op)
        # Defining internal set_background functions
        self.set_background1 = nl_op1.set_background
        self.set_background2 = nl_op2.set_background
        super(NonlinearVstack, self).__init__(self.nl_op, self.lin_op, self.set_background)
    
    def __str__(self):
        return "NLVstack"
    
    def set_background(self, model):
        """
        Set background function for the stack of Jacobian matrices
        """
        # Setting F(m0)
        self.set_background1(model)
        # Setting G(m0)
        self.set_background2(model)

# TODO add Hstack and Dstack


# Variable Projection Operator
class VarProOperator(Operator):
    """
        Operator of the form: h(m_nl)m_lin, for Variable-projection method
    """
    
    def __init__(self, h_nl, h_lin, set_nl, set_lin_jac, set_lin=None):
        """
            Constructor for an operator with a linear and non-linear model component
            Required arguments:
            h_nl    	= [no default] - non-linear operator class; Non-linear operator class
            h_lin   	= [no default] - operator class; Linear operator class
            set_nl  	= [no default] - class function pointer; Class function to set non-linear part within h_lin
            set_lin_jac = [no default] - class function pointer; Class function to set linear part within the Jacobian h_nl (if not necessary, use pyOperator.dummy_set_background)
            #Optional arguments:
            set_lin 	= [None] - class function pointer; Class function to set linear part within h_nl (not used during an inversion if ProblemL2VpReg is used)
        """
        if not isinstance(h_nl, NonlinearOperator):
            raise TypeError("ERROR! Not provided a non-linear operator class for h_nl")
        self.h_nl = h_nl
        self.h_lin = h_lin
        # Checking the range spaces
        if not h_nl.nl_op.range.checkSame(h_lin.range):
            raise ValueError("ERROR! The two provided operators have different ranges")
        self.set_nl = set_nl  # Function to set the non-linear component of the h(m_nl)
        self.set_lin_jac = set_lin_jac  # Function to set the non-linear component of the Jacobian H(m_nl;m_lin)
        self.set_lin = set_lin  # Function to set the non-linear component h(m_nl)m_lin
    
    def dotTest(self, verb=False, maxError=.0001):
        """
           Raising an exception, dot-product tests must be performed directly onto linear operator and the Jacobian of h(m_nl).
        """
        raise NotImplementedError(
            "ERROR! Perform dot-product tests directly onto linear operator and Jacobian of h(m_nl).")


# simple non-linear operator to tests linTest method
class cosOperator(Operator):
    """Cosine non-linear operator"""
    
    def __init__(self, domain):
        super(cosOperator, self).__init__(domain, domain)
    
    def forward(self, add, model, data):
        """Forward operator cos(x)"""
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        data.getNdArray()[:] += np.cos(model.getNdArray())
        return


class cosJacobian(Operator):
    """Jacobian of cosine non-linear operator (i.e., -sin(x0)dx)"""
    
    def __init__(self, domain):
        super(cosJacobian, self).__init__(domain, domain)
        self.background = domain.clone()
        self.backgroundNd = self.background.getNdArray()
    
    def forward(self, add, model, data):
        """Forward operator"""
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        data.getNdArray()[:] -= np.sin(self.backgroundNd) * model.getNdArray()
        return
    
    def adjoint(self, add, model, data):
        """Adjoint operator"""
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        model.getNdArray()[:] -= np.sin(self.backgroundNd) * data.getNdArray()
        return
    
    def set_background(self, background):
        """ Setting -sin(x0)"""
        self.background.copy(background)
        return



import numpy as np
import dask.distributed as daskD
from collections.abc import Iterable

from occamypy import Vector, Operator
from .vector import DaskVector
from .utils import DaskClient


def call_constructor(constr, args):
    """Function to call the constructor"""
    if isinstance(args, Iterable):
        op = constr(*args)
    else:
        op = constr(args)
    return op


def call_getDomain(opObj):
    """Function to call getDomain method"""
    domain = opObj.getDomain()
    return domain


def call_getRange(opObj):
    """Function to call getRange method"""
    range = opObj.getRange()
    return range


def call_forward(opObj, add, model, data):
    """Function to call forward operator"""
    res = opObj.forward(add, model, data)
    return res


def call_adjoint(opObj, add, model, data):
    """Function to call adjoint operator"""
    res = opObj.adjoint(add, model, data)
    return res


def getNdfuture(vecObj):
    """Function to obtain NdArray as a future object"""
    Nd = vecObj.getNdArray()
    return Nd


def call_func_name(opObj, func_name, *args):
    """Function to call a method by name"""
    fun2call = getattr(opObj, func_name)
    res = fun2call(*args)
    return res


def _add_from_NdArray(vecObj, NdArray):
    """Function to add vector values from numpy array"""
    vecObj.getNdArray()[:] += NdArray
    return


class DaskOperator(Operator):
    """
    Class to apply multiple operators in parallel through Dask and DaskVectors
    """
    
    def __init__(self, dask_client, op_constructor, op_args, chunks, **kwargs):
        """
        Dask Operator constructor
        
        :param dask_client: [no default] - DaskClient;
            client object to use when submitting tasks (see dask_util module)
        :param op_constructor: [no default] - pointer to function;
            Pointer to constructor
        :param op_args: [no default] - list;
            List containing lists of arguments to run the constructor.
            It can instantiate the same operator on multiple workers or different ones if requested
            by passing a list of list of arguments (e.g., [(arg1,arg2,arg3,...)])
        :param chunks: [no default] - list;
            List defininig how many operators wants to instantiated.
            Note, the list must contain the same number of elements as the number of
            Dask workers present in the DaskClient.
        :param setbackground_func_name: [None] - string;
            Name of the function to set the model point on which the Jacobian is computed.
            See NonLinearOperator in operator module.
        :param spread_op: [None] - DaskSpreadOp;
            Spreading operator to distribute a model vector to the set_background functions
        :param set_aux_name: [None] - string;
            Name of the function to set the auxiliary vector. Useful for VpOperator.
        :param spread_op_aux: [None] - DaskSpreadOp;
            Spreading operator to distribute an auxiliary vector to the set_aux functions
        """
        # Client to submit tasks
        if not isinstance(dask_client, DaskClient):
            raise TypeError("Passed client is not a Dask Client object!")
        if not isinstance(op_args, list):
            raise TypeError("Passed operator arguments not a list!")
        self.dask_client = dask_client
        self.client = self.dask_client.getClient()
        wrkIds = self.dask_client.getWorkerIds()
        N_wrk = self.dask_client.getNworkers()
        # Check if number of provided chunks is the same as workers
        if len(chunks) != N_wrk:
            raise ValueError(
                "Number of provide chunks (%s) different than the number of workers (%s)" % (len(chunks), N_wrk))
        # Check if many arguments are passed to construct different operators
        N_args = len(op_args)
        N_ops = int(np.sum(chunks))
        if N_args > 1:
            if N_args != N_ops:
                raise ValueError(
                    "Number of lists of arguments (%s) different than the number of requested operators (%s)" % (
                        N_args, N_ops))
        else:
            if N_ops > 1:
                op_args = [op_args for ii in range(N_ops)]
        
        # Instantiation of the operators on each worker
        self.dask_ops = []
        for iwrk, wrkId in enumerate(wrkIds):
            for iop in range(chunks[iwrk]):
                self.dask_ops.append(
                    self.client.submit(call_constructor,
                                       op_constructor,
                                       op_args.pop(0),
                                       workers=[wrkId],
                                       pure=False))
        daskD.wait(self.dask_ops)
        for idx in range(len(self.dask_ops)):
            if self.dask_ops[idx].status == 'error':
                print("Error for dask operator %s" % idx)
                print(self.dask_ops[idx].result())
        # Creating domain and range of the Dask operator
        dom_vecs = []  # List of remote domain vectors
        rng_vecs = []  # List of remote range vectors
        for op in self.dask_ops:
            dom_vecs.append(self.client.submit(call_getDomain, op, pure=False))
            rng_vecs.append(self.client.submit(call_getRange, op, pure=False))
        daskD.wait(dom_vecs + rng_vecs)
        self.domain = DaskVector(self.dask_client, dask_vectors=dom_vecs)
        self.range = DaskVector(self.dask_client, dask_vectors=rng_vecs)
        # Set background function name "necessary for non-linear operator Jacobian"
        self.set_background_name = kwargs.get("setbackground_func_name", None)
        if self.set_background_name:
            # Creating a spreading operator useful
            self.Sprd = kwargs.get("spread_op", None)
            if self.Sprd:
                if not isinstance(self.Sprd, DaskSpread):
                    raise TypeError("Provided spread_op not a DaskSpreadOp class!")
                self.model_tmp = self.Sprd.getRange().clone()
        # Set aux function name "necessary for VP operator"
        self.set_aux_name = kwargs.get("set_aux_name", None)
        if self.set_aux_name:
            # Creating a spreading operator useful
            self.SprdAux = kwargs.get("spread_op_aux", None)
            if self.SprdAux:
                if not isinstance(self.SprdAux, DaskSpread):
                    raise TypeError("Provided spread_op_aux not a DaskSpreadOp class!")
                self.tmp_aux = self.SprdAux.getRange().clone()
        return
    
    def __str__(self):
        return " DaskOp "
    
    def forward(self, add, model, data):
        """Forward Dask operator"""
        if not isinstance(model, DaskVector):
            raise TypeError("Model vector must be a DaskVector!")
        if not isinstance(data, DaskVector):
            raise TypeError("Data vector must be a DaskVector!")
        # Dimensionality check
        self.checkDomainRange(model, data)
        add = [add] * len(self.dask_ops)
        fwd_ftr = self.client.map(call_forward, self.dask_ops, add, model.vecDask, data.vecDask, pure=False)
        daskD.wait(fwd_ftr)
        return
    
    def adjoint(self, add, model, data):
        """Adjoint Dask operator"""
        if not isinstance(model, DaskVector):
            raise TypeError("Model vector must be a DaskVector!")
        if not isinstance(data, DaskVector):
            raise TypeError("Data vector must be a DaskVector!")
        # Dimensionality check
        self.checkDomainRange(model, data)
        add = [add] * len(self.dask_ops)
        adj_ftr = self.client.map(call_adjoint, self.dask_ops, add, model.vecDask, data.vecDask, pure=False)
        daskD.wait(adj_ftr)
        return
    
    def set_background(self, model):
        """Function to call set_background function of each dask operator"""
        if self.set_background_name == None:
            raise NameError("setbackground_func_name was not defined when constructing the operator!")
        if self.Sprd:
            self.Sprd.forward(False, model, self.model_tmp)
            model = self.model_tmp
        setbkg_ftr = self.client.map(call_func_name,
                                     self.dask_ops,
                                     [self.set_background_name] * self.dask_client.getNworkers(),
                                     model.vecDask,
                                     pure=False)
        daskD.wait(setbkg_ftr)
        return
    
    def set_aux(self, aux_vec):
        """Function to call set_nl or set_lin_jac functions of each dask operator"""
        if self.set_aux_name == None:
            raise NameError("set_aux_name was not defined when constructing the operator!")
        if self.SprdAux:
            self.SprdAux.forward(False, aux_vec, self.tmp_aux)
            aux_vec = self.tmp_aux
        setaux_ftr = self.client.map(call_func_name,
                                     self.dask_ops,
                                     [self.set_aux_name] * self.dask_client.getNworkers(),
                                     aux_vec.vecDask,
                                     pure=False)
        daskD.wait(setaux_ftr)
        return


class DaskSpread(Operator):
    """
    Class to spread/stack single vector to/from multiple copies on different workers:
         | v1 |   | I |
    fwd: | v2 | = | I | v       adj: | v | = | I | v1 + | I | v2 + | I | v3
         | v3 |   | I |
    """
    
    def __init__(self, dask_client, domain, chunks):
        """
        Dask Operator constructor
        
        :param dask_client: [no default] - DaskClient;
            client object to use when submitting tasks (see dask_util module)
        :param domain: [no default] - vector class;
            Vector template to be spread/stack (note this is also the domain of the operator)
        :param chunks: [no default] - list;
            List defininig how many operators wants to instantiated.
            It must have the same number of elements as the number of Dask workers present in the DaskClient.
        """
        if not isinstance(dask_client, DaskClient):
            raise TypeError("Passed client is not a Dask Client object!")
        if not isinstance(domain, Vector):
            raise TypeError("domain is not a vector-derived object!")
        self.dask_client = dask_client
        self.client = self.dask_client.getClient()
        self.chunks = chunks
        self.setDomainRange(domain, DaskVector(self.dask_client, vector_template=domain, chunks=chunks))
        return
    
    def __str__(self):
        return "DaskSprd"
    
    def forward(self, add, model, data):
        """Forward operator"""
        if not isinstance(data, DaskVector):
            raise TypeError("Data vector must be a DaskVector!")
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        # Model vector checking
        if isinstance(model, DaskVector):
            # Getting the future to the first vector in the Dask vector
            modelNd = self.client.submit(getNdfuture, model.vecDask[0], pure=False)
        else:
            # Getting the numpy array to the local model vector
            modelNd = model.getNdArray()
        
        # Spreading model array to workers
        if len(self.chunks) == self.dask_client.getNworkers():
            dataVecList = data.vecDask.copy()
            for iwrk, wrkId in enumerate(self.dask_client.getWorkerIds()):
                arrD = self.client.scatter(modelNd, workers=[wrkId])
                daskD.wait(arrD)
                for ii in range(self.chunks[iwrk]):
                    daskD.wait(
                        self.client.submit(_add_from_NdArray, dataVecList.pop(0), arrD, workers=[wrkId], pure=False))
        else:
            # Letting Dask handling the scattering of the data (not ideal)
            futures = self.client.map(_add_from_NdArray,
                                      data.vecDask,
                                      [modelNd] * len(data.vecDask),
                                      pure=False)
            daskD.wait(futures)
        return
    
    def adjoint(self, add, model, data):
        """Adjoint operator"""
        if not isinstance(data, DaskVector):
            raise TypeError("Data vector must be a DaskVector!")
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        arrD = self.client.map(getNdfuture, data.vecDask, pure=False)
        daskD.wait(arrD)
        sum_array = self.client.submit(np.sum, arrD, axis=0, pure=False)
        daskD.wait(sum_array)
        if isinstance(model, DaskVector):
            # Getting the future to the first vector in the Dask vector
            daskD.wait(self.client.submit(_add_from_NdArray, model, sum_array, pure=False))
        else:
            # Getting the numpy array to the local model vector
            modelNd = model.getNdArray()
            modelNd[:] += sum_array.result()
        return


class DaskCollect(Operator):
    """
    Class to Collect/Scatter a Dask vector into/from a local vector
    """
    
    def __init__(self, domain, range):
        """
        Dask Collect constructor
        
        :param domain : - DaskVector;
            Dask vector to be collected from remote
        :param range : - Vector;
            Vector class to be locally stored
        """
        if not isinstance(domain, DaskVector):
            raise TypeError("domain is not a DaskVector object!")
        if isinstance(range, DaskVector):
            raise TypeError("range should not a DaskVector object!")
        if not isinstance(range, Vector):
            raise TypeError("range is not a vector-derived object!")
        if domain.size != range.size:
            raise ValueError("number of elements in domain and range is not equal!")
        super(DaskCollect, self).__init__(domain, range)
    
    def forward(self, add, model, data):
        """Forward operator: collecting dask vector array to local one"""
        if not isinstance(model, DaskVector):
            raise TypeError("Model vector must be a DaskVector!")
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        dataNd = data.getNdArray().ravel()
        # Obtaining remove array/s
        modelNd_list = model.getNdArray()
        idx = 0
        # Adding remove arrays to local one
        for arr in modelNd_list:
            n_elem = arr.size
            dataNd[idx:idx + n_elem] += arr.ravel()
            idx += n_elem
        return
    
    def adjoint(self, add, model, data):
        """Adjoint operator: scattering/distributing local array to remove vector"""
        if not isinstance(model, DaskVector):
            raise TypeError("Model vector must be a DaskVector!")
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        dataNd = data.getNdArray().ravel()
        # Shapes of the DaskVector chunks
        shapes = model.shape
        # Scattering local array
        client = model.client
        idx_el = 0
        for idx, shape in enumerate(shapes):
            n_elem = np.prod(shape)
            wrkId = list(client.who_has(model.vecDask[idx]).values())[0]
            arrD = client.scatter(dataNd[idx_el:idx_el + n_elem], workers=wrkId)
            daskD.wait(arrD)
            daskD.wait(client.submit(_add_from_NdArray, model.vecDask[idx], arrD, pure=False))
            idx_el += n_elem
        return



from math import isnan
import numpy as np

from occamypy import operator as O
from occamypy import vector as V
from occamypy import problem as P
from occamypy.solver import Solver, BasicStopper, CG, LSQR, SD
from occamypy.utils import Logger

zero = 10 ** (np.floor(np.log10(np.abs(float(np.finfo(np.float64).tiny)))) + 2)  # Check for avoid Overflow or Underflow


def _soft_thresh(x, thresh):
    """
    Soft-thresholding function:
        y = sign(x) * max(abs(x) - thresh, 0)

    :param x        : vector, input values
    :param thresh   : float, soft threshold
    :return         : vector, output clipped values
    """
    return x.clone().sign() * x.clone().abs().addbias(-thresh).maximum(0.)


def _shrinkage(x, thresh, eps=1e-10):
    """
    Shrinkage function Gamma
        y = x / (|x| + eps) * maximum(|x| - thresh, 0)

    note: sgn(x) = x / |x| so it is the same of _soft_thresh
    """
    y = x.clone()
    y.multiply(x.clone().abs().addbias(eps).reciprocal())
    return y * x.clone().abs().addbias([-t for t in thresh]).maximum(0.)


def _shrinkage1(x, thresh):
    xabs = np.abs(x)
    return x / (xabs + 1e-10) * np.maximum(xabs - thresh, 0)


def _proximal_L2(x, thresh, eps=1e-10):
    """
    Proximal operator for L2 distance as implemented in
    https://sporco.readthedocs.io/en/latest/modules/sporco.prox.html#sporco.prox.prox_l2
    """
    x_norm = x.norm()
    m = max(0, x_norm - thresh)
    c = m / (x_norm + eps)
    return x.clone().scale(c)


class ISTA(Solver):
    """
    Iterative Shrikage-Thresholding Algorithm (ISTA) solver to solve:
        1/2*| y - Am |_2 + lambda*| m |_1
    """
    
    def __init__(self, stopper, fast=False, logger=None):
        """
        Constructor for ISTA Solver:
        :param stopper: Stopper, object to terminate inversion
        :param fast: bool, apply the Fast-ISTA [False]
        :param logger: Logger, object to write inversion log file
        """
        # Calling parent construction
        super(ISTA, self).__init__()
        # Defining stopper object
        self.stopper = stopper
        # Logger object to write on log file
        self.logger = logger
        # Overwriting logger of the Stopper object
        self.stopper.logger = self.logger
        # Setting the fast flag
        self.fast = fast
        # print formatting
        self.iter_msg = "iter = %s, obj = %.5e, resnorm = %.2e, gradnorm = %.2e, feval = %d"
    
    def __del__(self):
        """Default destructor"""
        return
    
    def run(self, problem, verbose=False, restart=False):
        """Running ISTA solver"""
        
        self.create_msg = verbose or self.logger
        # Resetting stopper before running the inversion
        self.stopper.reset()
        # Checking if the provided problem is L1-LASSO
        if not isinstance(problem, P.Lasso):
            raise TypeError("Provided inverse problem not ProblemL1Lasso!")
        # Checking if the regularization weight was set
        if problem.lambda_value is None:
            raise ValueError("Regularization weight (lambda_value) is not set!")
        
        if not restart:
            if self.create_msg:
                msg = 90 * "#" + "\n"
                msg += "\t\t\t\tFAST " if self.fast else "\t\t\t\t\t"
                msg += "ITERATIVE SHRINKAGE-THRESHOLDING ALGORITHM log file\n"
                msg += "\tRestart folder: %s\n" % self.restart.restart_folder
                msg += "\tModeling Operator:\t\t%s\n" % problem.op
                msg += "\tRegularization weight:\t%.2e\n" % problem.lambda_value
                msg += 90 * "#" + "\n"
                if verbose:
                    print(msg.replace(" log file", ""))
                if self.logger:
                    self.logger.addToLog(msg)
            
            # Setting internal vectors (model, search direction, and previous gradient vectors)
            prblm_mdl = problem.get_model()
            ista_mdl = prblm_mdl.clone()
            # Other parameters in case FISTA is requested
            if self.fast:
                t = 1.0
                # fista_mdl = prblm_mdl.clone()
            
            # Other internal variables
            iiter = 0
        else:
            # Retrieving parameters and vectors to restart the solver
            if self.create_msg:
                msg = "Restarting previous solver run from: %s" % self.restart.restart_folder
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
            self.restart.read_restart()
            # Retrieving inversion parameters
            iiter = self.restart.retrieve_parameter("iter")
            initial_obj_value = self.restart.retrieve_parameter("obj_initial")
            ista_mdl = self.restart.retrieve_vector("ista_mdl")
            # Other parameters in case FISTA is requested
            if self.fast:
                t = self.restart.retrieve_parameter("t")
                # fista_mdl = self.restart.retrieve_vector("fista_mdl")
        
        ista_mdl0 = ista_mdl.clone()  # Previous model in case stepping procedure fails
        
        # Inversion loop
        while True:
            obj0 = problem.get_obj(ista_mdl)  # Compute objective function value
            prblm_grad = problem.get_grad(ista_mdl)  # Compute the gradient g = - A' [y - Ax]
            if iiter == 0:
                # Saving initial objective function value
                initial_obj_value = obj0
                self.restart.save_parameter("obj_initial", initial_obj_value)
                if self.create_msg:
                    msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                           obj0,
                                           problem.get_rnorm(ista_mdl),
                                           problem.get_gnorm(ista_mdl),
                                           problem.get_fevals())
                    # Writing on log file
                    if verbose:
                        print(msg)
                    if self.logger:
                        self.logger.addToLog(msg)
                # Check if either objective function value or gradient norm is NaN
                if isnan(obj0) or isnan(prblm_grad.norm()):
                    raise ValueError("Either gradient norm or objective function value NaN!")
            if problem.get_gnorm(ista_mdl) == 0.:
                print("Gradient vanishes identically")
                break
            
            # Saving results
            self.save_results(iiter, problem, force_save=False)
            ista_mdl0.copy(ista_mdl)  # Saving model before updating it
            
            # Update model x = x + scale_precond * A' [y - Ax]
            ista_mdl.scaleAdd(prblm_grad, 1.0, -1.0 / problem.op_norm)
            
            # SOFT-THRESHOLDING STEP
            ista_mdl.copy(_soft_thresh(ista_mdl, problem.lambda_value / problem.op_norm))
            
            # Projecting model onto the bounds (if any)
            if "bounds" in dir(problem):
                problem.bounds.apply(ista_mdl)
            
            if self.fast:
                t0 = t
                t = (1. + np.sqrt(1. + 4. * t ** 2)) / 2.
                # z = x + ((t0 - 1.) / t) * (x - xold)
                scale = (t0 - 1.) / t
                ista_mdl.scaleAdd(ista_mdl0, 1.0 + scale, -scale)
            
            obj1 = problem.get_obj(ista_mdl)
            if obj1 >= obj0:
                if self.create_msg:
                    msg = "Objective function didn't reduce, will terminate solver:\n\t" \
                          "obj_new = %.2e\tobj_cur = %.2e" % (obj1, obj0)
                    if verbose:
                        print(msg)
                    # Writing on log file
                    if self.logger:
                        self.logger.addToLog(msg)
                # Copying back to the previous solution
                ista_mdl.copy(ista_mdl0)
                break
            
            # Saving current model in case of restart and other parameters
            self.restart.save_parameter("iter", iiter)
            self.restart.save_vector("ista_mdl", ista_mdl)
            if self.fast:
                self.restart.save_parameter("t", t)
                # self.restart.save_vector("fista_mdl", fista_mdl)
            
            # iteration info
            iiter = iiter + 1
            if self.create_msg:
                msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                       obj1,
                                       problem.get_rnorm(ista_mdl),
                                       problem.get_gnorm(ista_mdl),
                                       problem.get_fevals())
                if verbose:
                    print(msg)
                # Writing on log file
                if self.logger:
                    self.logger.addToLog("\n" + msg)
            # Check if either objective function value or gradient norm is NaN
            if isnan(obj1) or isnan(prblm_grad.norm()):
                raise ValueError("Either gradient norm or objective function value NaN!")
            if self.stopper.run(problem, iiter, initial_obj_value, verbose):
                break
        
        # Writing last inverted model
        self.save_results(iiter, problem, force_save=True, force_write=True)
        if self.create_msg:
            msg = 90 * "#" + "\n"
            msg += "\t\t\t\tFAST " if self.fast else "\t\t\t\t\t"
            msg += "ITERATIVE SHRINKAGE-THRESHOLDING ALGORITHM log file end\n"
            msg += 90 * "#" + "\n"
            if verbose:
                print(msg.replace(" log file", ""))
            if self.logger:
                self.logger.addToLog(msg)
        # Clear restart object
        self.restart.clear_restart()


class ISTC(Solver):
    """ISTC solver to solve: convex problem 1/2*| y - Am |_2 + lambda*| m |_1"""
    
    def __init__(self, stopper, inner_it, cooling_start, cooling_end, logger=None):
        """
        Constructor for ISTC Solver
        :param stopper      : Stopper, object to terminate inversion
        :param inner_it     : int, Number of inner iterations
        :param logger       : Logger, object to write inversion log file
        :param cooling_start: float, Start of cooling continuation as fraction of size of sorted array |A'y|
        :param cooling_end  : float; End of cooling continuation as fraction of size of sorted array |A'y|
        """
        # Calling parent construction
        super(ISTC, self).__init__()
        # Defining stopper object
        self.stopper = stopper
        # Logger object to write on log file
        self.logger = logger
        # Overwriting logger of the Stopper object
        self.stopper.logger = self.logger
        self.iter_msg = "Inner_iter = %s, obj = %.5e, resnorm = %.2e, gradnorm= %.2e, feval = %d"
        
        # ISTC parameters
        if self.stopper.niter <= 0:
            raise ValueError("niter for stopper object must be positive and greater than 0!")
        self.inner_it = inner_it  # number of inner iterations, the outer iterations are taken care by the stopper
        # cooling_start and cooling_end are numbers between 0 and 1 such that cooling_start <= cooling_end
        if not 0 <= cooling_start <= 1 or not 0 <= cooling_end <= 1 or cooling_end < cooling_start:
            raise ValueError("Cooling_start and end must be within [0,1] interval and cooling_start <= cooling_end")
        self.cooling_start = cooling_start  # start of cooling continuation as fraction of size of sorted array |A'y|
        self.cooling_end = cooling_end  # end of cooling continuation as fraction of size of sorted array |A'y|
    
    def __del__(self):
        """Default destructor"""
        return
    
    def run(self, problem, verbose=False, restart=False):
        """Running ISTC solver"""
        
        self.create_msg = verbose or self.logger
        
        # Resetting stopper before running the inversion
        self.stopper.reset()
        # Checking if the provided problem is L1-LASSO
        if not isinstance(problem, P.Lasso):
            raise TypeError("Provided inverse problem not ProblemL1Lasso!")
        # Computing preconditioning
        scale_precond = 0.99 * np.sqrt(2) / problem.op_norm  # scaling factor applied to operator A for preconditioning
        if not restart:
            if self.create_msg:
                msg = 90 * "#" + "\n"
                msg += "\t\t\tITERATIVE SOFT-THRESHOLDING WITH COOLING SOLVER log file\n"
                msg += "\tRestart folder: %s\n" % self.restart.restart_folder
                msg += "\tModeling Operator:\t\t%s\n" % problem.op
                msg += "\tRegularization weight:\t%.2e\n" % problem.lambda_value
                msg += 90 * "#" + "\n"
                if verbose:
                    print(msg.replace(" log file", ""))
                if self.logger:
                    self.logger.addToLog(msg)
            
            # Setting internal vectors (model, search direction, and previous gradient vectors)
            prblm_mdl = problem.get_model()
            istc_mdl = prblm_mdl.clone()
            
            # Inversion always starts from m = 0
            # (I need to understand if it is possible to start from m different than 0)
            istc_mdl.zero()  # modl = 0
            # Other internal variables
            iiter = 0
            # Computing cooling schedule for lambda values
            # istc_mdl.scale(scale_precond) #Currently unnecessary since starting model is zero
            prblm_grad = problem.get_grad(istc_mdl)
            grad_arr = np.copy(prblm_grad.getNdArray())
            grad_arr = np.abs(grad_arr.flatten())  # |A'y| and removing zero elements
            grad_arr = grad_arr[np.nonzero(grad_arr)]
            if grad_arr.size == 0:
                raise ValueError("-- A'y is returning a null vector (i.e., y in the Null space of A')")
            # Sorting the elements in descending order
            grad_arr.sort()
            grad_arr = np.flip(grad_arr, 0)
            # Setting fraction of points sampled by the outer loop (linear sampling)
            samples = np.array(np.round(np.linspace(self.cooling_start,
                                                    self.cooling_end,
                                                    self.stopper.niter)
                                        * grad_arr.size), dtype=np.uint64, copy=False)
            # Lambda values to be used during inversion for each outer loop iteration
            lambda_values = grad_arr[samples]
            # Scaling by the preconditioning
            lambda_values *= scale_precond
            # Saving the lambda values to avoid recomputation if restart is used
            self.restart.save_parameter("lambda_values", lambda_values)
        else:
            # Retrieving parameters and vectors to restart the solver
            if self.create_msg:
                msg = "Restarting previous solver run from: %s" % self.restart.restart_folder
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
            self.restart.read_restart()
            # Retrieving lambda values and other parameters
            lambda_values = self.restart.retrieve_parameter("lambda_values")
            iiter = self.restart.retrieve_parameter("iter")
            initial_obj_value = self.restart.retrieve_parameter("obj_initial")
            istc_mdl = self.restart.retrieve_vector("istc_mdl")
        
        # Common variables unrelated to restart
        success = True
        istc_mdl0 = istc_mdl.clone()  # Previous model in case stepping procedure fails
        istc_mdl_save = istc_mdl0  # used also to save results
        
        # Outer iteration loop
        while True:
            # Setting lambda value for a given outer loop iteration
            problem.set_lambda(lambda_values[iiter])
            problem.obj_updated = False  # Lambda has been changed so objective function will change as well
            if self.create_msg:
                msg = "Outer_iter = %s\tlambda_value = %.2e" % (
                    str(iiter).zfill(self.stopper.zfill), lambda_values[iiter])
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
            if not restart:
                inner_iter = 0
            else:
                self.restart.retrieve_parameter("inner_iter", inner_iter)
                restart = False
            
            if iiter == 0:
                # Applying preconditioning
                istc_mdl.scale(scale_precond)
                obj = problem.get_obj(istc_mdl)  # Compute objective function value
                # Saving initial objective function value
                initial_obj_value = obj
                self.restart.save_parameter("obj_initial", initial_obj_value)
            while inner_iter < self.inner_it:
                obj0 = problem.get_obj(istc_mdl)  # Compute objective function value
                prblm_grad = problem.get_grad(istc_mdl)  # Compute the gradient g = - A' [y - Ax]
                if inner_iter == 0:
                    if self.create_msg:
                        msg = self.iter_msg % (str(inner_iter).zfill(self.stopper.zfill),
                                               obj0,
                                               problem.get_rnorm(istc_mdl),
                                               problem.get_gnorm(istc_mdl),
                                               problem.get_fevals())
                        # Writing on log file
                        if verbose:
                            print(msg)
                        if self.logger:
                            self.logger.addToLog(msg)
                    # Check if either objective function value or gradient norm is NaN
                    if isnan(obj0) or isnan(prblm_grad.norm()):
                        raise ValueError("Either gradient norm or objective function value NaN!")
                if problem.get_gnorm(istc_mdl) == 0.:
                    print("Gradient vanishes identically")
                    break
                
                # Removing preconditioning scaling factor from inverted model
                istc_mdl_save.copy(istc_mdl)
                istc_mdl_save.scale(scale_precond)
                # Saving results
                self.save_results(iiter, problem, model=istc_mdl_save, force_save=False)
                
                # Stepping for internal iteration model update
                istc_mdl0.copy(istc_mdl)  # Saving model before updating it
                istc_mdl.scaleAdd(prblm_grad, 1.0, -scale_precond)  # Update model x = x + scale_precond * A' [y - Ax]
                #########################################
                # SOFT-THRESHOLDING STEP
                istc_mdl = _soft_thresh(istc_mdl, problem.lambda_value)
                #########################################
                # Projecting model onto the bounds (if any)
                if "bounds" in dir(problem):
                    problem.bounds.apply(istc_mdl)
                
                obj1 = problem.get_obj(istc_mdl)
                problem.get_model().writeVec("problem_model.H")
                istc_mdl.writeVec("solver_model.H")
                if obj1 >= obj0:
                    if self.create_msg:
                        msg = "Objective function didn't reduce, will terminate solver:\n\t" \
                              "obj_new = %.2e\tobj_cur = %.2e" % (obj1, obj0)
                        if verbose:
                            print(msg)
                        # Writing on log file
                        if self.logger:
                            self.logger.addToLog(msg)
                    # Copying back to the previous solution
                    istc_mdl.copy(istc_mdl0)
                    break
                
                # Saving current model in case of restart and other parameters
                self.restart.save_parameter("iter", iiter)
                self.restart.save_parameter("inner_iter", inner_iter)
                self.restart.save_vector("istc_mdl", istc_mdl)
                
                # iteration info
                inner_iter += 1
                if self.create_msg:
                    msg = self.iter_msg % (str(inner_iter).zfill(self.stopper.zfill),
                                           obj1,
                                           problem.get_rnorm(istc_mdl),
                                           problem.get_gnorm(istc_mdl),
                                           problem.get_fevals())
                    if verbose:
                        print(msg)
                    # Writing on log file
                    if self.logger:
                        self.logger.addToLog(msg)
                # Check if either objective function value or gradient norm is NaN
                if isnan(obj1) or isnan(prblm_grad.norm()):
                    raise ValueError("Either gradient norm or objective function value NaN!")
            iiter = iiter + 1
            if self.stopper.run(problem, iiter, initial_obj_value, verbose):
                break
        
        # Removing preconditioning scaling factor from inverted model
        istc_mdl_save.copy(istc_mdl)
        istc_mdl_save.scale(scale_precond)
        # Writing last inverted model
        self.save_results(iiter, problem, model=istc_mdl_save, force_save=True, force_write=True)
        if self.create_msg:
            msg = 90 * "#" + "\n"
            msg += "\t\t\tITERATIVE SOFT-THRESHOLDING WITH COOLING SOLVER log file end\n"
            msg += 90 * "#" + "\n"
            if verbose:
                print(msg.replace(" log file", ""))
            if self.logger:
                self.logger.addToLog(msg)
        # Clear restart object
        self.restart.clear_restart()


class SplitBregman(Solver):
    """Split-Bregman solver for GeneralizedLasso problems"""
    
    # Default class methods/functions
    def __init__(self, stopper, logger=None, niter_inner=3, niter_solver=5, breg_weight=1., linear_solver='CG',
                 warm_start=False, mod_tol=1e-10):
        """
        Constructor for Split-Bregman Solver
        :param stopper          : stopper object
        :param logger           : logger object
        :param niter_inner      : int; number of iterations for the shrinkage loop [default 3]
        :param niter_solver     : int; number of iterations for the internal linear solver [default 5]
        :param breg_weight      : float; coefficient for the Bregman update b += beta * (R*x - d) [1.]
        :param linear_solver    : str; linear solver to be used [CG, SD, LSQR]
        :param warm_start       : bool; linear solver restarts from previous solution of inner problem [False]
        :param mod_tol          : float; stop criterion for relative change of model norm
        """
        # Calling parent construction
        super(SplitBregman, self).__init__()
        # Defining stopper object
        self.stopper = stopper
        # Logger object to write on log file
        self.logger = logger
        # Overwriting logger of the Stopper object
        self.stopper.logger = self.logger
        # Model norm change stop criterion
        self.mod_tol = mod_tol
        
        # Logger for internal linear solver
        self.logger_lin_solv = None
        if logger is not None:
            if "/" in logger.file.name:
                folder = "/".join(logger.file.name.split("/")[:-1]) + "/"
            else:
                folder = ""
            filename = "inner_inv_" + logger.file.name.split("/")[-1]
            self.logger_lin_solv = Logger(folder + filename)
        
        self.niter_inner = niter_inner  # number of iterations for the shrinkage
        self.niter_solver = niter_solver  # number of iterations for the internal problem
        self.warm_start = warm_start
        if breg_weight > 1.:
            raise ValueError("ERROR! Bregman update weight has to be <= 1")
        self.breg_weight = float(abs(breg_weight))
        
        if linear_solver == 'CG':
            self.linear_solver = CG(BasicStopper(niter=self.niter_solver), logger=self.logger_lin_solv)
        elif linear_solver == 'SD':
            self.linear_solver = SD(BasicStopper(niter=self.niter_solver), logger=self.logger_lin_solv)
        elif linear_solver == 'LSQR':
            self.linear_solver = LSQR(BasicStopper(niter=self.niter_solver), logger=self.logger_lin_solv)
        else:
            raise ValueError('ERROR! Solver has to be CG, SD or LSQR')
        self.linear_solver.setDefaults(iter_sampling=1, flush_memory=True)
        
        # print formatting
        self.iter_msg = "iter = %s, obj = %.5e, df_obj = %.2e, reg_obj = %.2e, resnorm = %.2e"
    
    def run(self, problem, verbose=False, inner_verbose=False, restart=False):
        """Running SplitBregman solver"""
        if type(problem) != P.GeneralizedLasso:
            raise TypeError("Input problem object must be a GeneralizedLasso")
        
        verbose = True if inner_verbose else verbose
        self.create_msg = verbose or self.logger
        
        # overriding save_grad variable
        self.save_grad = False
        
        # reset stopper before running the inversion
        self.stopper.reset()
        
        # initialize all the vectors and operators for Split-Bregman
        breg_b = problem.reg_op.range.clone().zero()
        breg_d = breg_b.clone()
        RL1x = breg_b.clone()  # store RegL1 * solution
        
        sb_mdl = problem.model.clone()
        if sb_mdl.norm() != 0.:
            self.warm_start = True
        sb_mdl_old = problem.model.clone()
        
        reg_op = np.sqrt(problem.eps) * problem.reg_op   # TODO can we avoid this?
        
        # inner problem
        linear_problem = P.LeastSquares(model=sb_mdl.clone(),
                                        data=V.superVector(problem.data, breg_d.clone()),
                                        op=O.Vstack(problem.op, reg_op),
                                        minBound=problem.minBound,
                                        maxBound=problem.maxBound,
                                        boundProj=problem.boundProj)
        
        if restart:
            self.restart.read_restart()
            outer_iter = self.restart.retrieve_parameter("iter")
            initial_obj_value = self.restart.retrieve_parameter("obj_initial")
            sb_mdl = self.restart.retrieve_vector("sb_mdl")
            if self.create_msg:
                msg = "Restarting previous solver run from: %s" % self.restart.restart_folder
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
        
        else:
            outer_iter = 0
            if self.create_msg:
                msg = 90 * '#' + '\n'
                msg += "\t\t\tSPLIT-BREGMAN ALGORITHM log file\n\n"
                msg += "\tRestart folder: %s\n" % self.restart.restart_folder
                msg += "\tModeling Operator:\t%s\n" % problem.op
                msg += "\tInner iterations:\t%d\n" % self.niter_inner
                msg += "\tSolver iterations:\t%d\n" % self.niter_solver
                msg += "\tL1 Regularizer op:\t%s\n"  % problem.reg_op
                msg += "\tL1 Regularizer weight:\t%.2e\n" % problem.eps
                msg += "\tBregman update weight:\t%.2e\n" % self.breg_weight
                if self.warm_start:
                    msg += "\tUsing warm start option for inner problem\n"
                msg += 90 * '#' + '\n'
                if verbose:
                    print(msg.replace(" log file", ""))
                if self.logger:
                    self.logger.addToLog(msg)
                if self.logger_lin_solv:
                    msg = 90 * '#' + '\n'
                    msg += "\t\t\tSPLIT-BREGMAN ALGORITHM internal inversions log file\n"
                    msg += 90 * '#' + '\n'
                    self.logger_lin_solv.addToLog(msg)
        
        # Main iteration loop
        while True:
            obj0 = problem.get_obj(sb_mdl)
            # Saving previous model vector
            sb_mdl_old.copy(sb_mdl)
            
            if outer_iter == 0:
                initial_obj_value = obj0
                self.restart.save_parameter("obj_initial", initial_obj_value)
                if self.create_msg:
                    msg = self.iter_msg % (str(outer_iter).zfill(self.stopper.zfill),
                                           obj0,
                                           problem.obj_terms[0],
                                           obj0 - problem.obj_terms[0],
                                           problem.get_rnorm(sb_mdl))
                    if verbose:
                        print(msg)
                    if self.logger:
                        self.logger.addToLog("\n" + msg)
            
            if self.logger_lin_solv:
                self.logger_lin_solv.addToLog("\n\t\t\tOuter iteration: %s"
                                              % (str(outer_iter).zfill(self.stopper.zfill)))
            
            if isnan(obj0):
                raise ValueError("Objective function values NaN!")
            
            if obj0 <= zero:
                print("Objective function is numerically zero! Stop the inversion")
                break
            
            self.save_results(outer_iter, problem, force_save=False)
            
            for iter_inner in range(self.niter_inner):
                
                if self.logger_lin_solv:
                    msg = "\t\tstarting inner iter %d with d = %.2e, b = %.2e" \
                          % (iter_inner, breg_d.norm(), breg_b.norm())
                    self.logger_lin_solv.addToLog("\n" + msg)
                
                # resetting inversion problem variables
                if not self.warm_start:
                    linear_problem.model.zero()
                # prior = d - b
                linear_problem.data.vecs[-1].copy(breg_b)
                linear_problem.data.vecs[-1].scaleAdd(breg_d, -1., 1.)
                linear_problem.data.vecs[-1].scale(np.sqrt(problem.eps))  # TODO can we avoid this?
                linear_problem.setDefaults()
                
                # solve inner problem
                self.linear_solver.run(linear_problem, verbose=inner_verbose)
                
                # compute RL1*x
                problem.reg_op.forward(False, linear_problem.model, RL1x)
                
                # update breg_d
                breg_d.copy(_soft_thresh(RL1x.clone() + breg_b, thresh=problem.eps))
                
                if self.logger_lin_solv:
                    msg = "\t\tfinished inner iter %d with sb_mdl = %.2e, RL1x = %.2e" \
                          % (iter_inner, linear_problem.model.norm(), RL1x.norm())
                    self.logger_lin_solv.addToLog(msg)
            
            # update breg_b
            breg_b.scaleAdd(RL1x, 1.0, self.breg_weight)
            breg_b.scaleAdd(breg_d, 1., -self.breg_weight)
            
            # Update SB model
            sb_mdl.copy(linear_problem.model)
            
            # check objective function
            obj1 = problem.get_obj(sb_mdl)
            sb_mdl_norm = sb_mdl.norm()
            chng_norm = sb_mdl_old.scaleAdd(sb_mdl, 1., -1.).norm()
            if chng_norm <= self.mod_tol * sb_mdl_norm:
                if self.create_msg:
                    msg = "Relative model change (%.4e) norm smaller than given tolerance (%.4e)"\
                          % (chng_norm, self.mod_tol * sb_mdl_norm)
                    if verbose:
                        print(msg)
                    if self.logger:
                        self.logger.addToLog(msg)
                break
            
            # iteration info
            outer_iter += 1
            if self.create_msg:
                msg = self.iter_msg % (str(outer_iter).zfill(self.stopper.zfill),
                                       obj1,
                                       problem.obj_terms[0],
                                       obj1 - problem.obj_terms[0],
                                       problem.get_rnorm(sb_mdl))
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog("\n" + msg)
            
            # saving in case of restart
            self.restart.save_parameter("iter", outer_iter)
            self.restart.save_vector("sb_mdl", sb_mdl)
            
            if self.stopper.run(problem, outer_iter, initial_obj_value, verbose):
                break
        
        # writing last inverted model
        self.save_results(outer_iter, problem, force_save=True, force_write=True)
        
        # ending message and log file
        if self.create_msg:
            msg = 90 * '#' + '\n'
            msg += "\t\t\tSPLIT-BREGMAN ALGORITHM log file end\n"
            msg += 90 * '#'
            if verbose:
                print(msg.replace(" log file", ""))
            if self.logger:
                self.logger.addToLog("\n" + msg)
        
        # Clear restart object
        self.restart.clear_restart()



import numpy as np
import os
from time import time
from copy import deepcopy
from shutil import copyfile

from .base import Vector
from occamypy.utils import sep
from occamypy.utils.os import RunShellCmd, hashfile, BUF_SIZE

from re import compile
re_dpr = compile("DOT RESULT(.*)")


class VectorOC(Vector):
    """Out-of-core python vector class"""
    
    def __init__(self, in_content):
        """VectorOC constructor: input= numpy array, header file, vectorIC"""
        # Verify that input is a numpy array or header file or vectorOC
        super(VectorOC).__init__()
        if isinstance(in_content, Vector):
            # VectorIC passed to constructor
            # Placing temporary file into datapath folder
            tmp_vec = sep.datapath + "tmp_vectorOC" + str(int(time() * 1000000)) + ".H"
            sep.write_file(tmp_vec, in_content.getNdArray(), in_content.ax_info)
            self.vecfile = tmp_vec  # Assigning internal vector array
            # Removing header file? (Default behavior is to remove temporary file)
            self.remove_file = True
        elif isinstance(in_content, np.ndarray):
            # Numpy array passed to constructor
            tmp_vec = sep.datapath + "tmp_vectorOC" + str(int(time() * 1000000)) + ".H"
            sep.write_file(tmp_vec, in_content)
            self.vecfile = tmp_vec  # Assigning internal vector array
            # Removing header file? (Default behavior is to remove temporary file)
            self.remove_file = True
        elif isinstance(in_content, str):
            # Header file passed to constructor
            self.vecfile = in_content  # Assigning internal vector array
            # Removing header file? (Default behavior is to preserve user file)
            self.remove_file = False
        else:
            # Not supported type
            raise ValueError("ERROR! Input variable not currently supported!")
        # Assigning binary file pointer
        self.binfile = sep.get_binary(self.vecfile)
        # Number of axes integer
        self.ndim = sep.get_num_axes(self.vecfile)
        # Number of elements per axis (tuple)
        axes_info = sep.get_axes(self.vecfile)
        axis_elements = tuple([ii[0] for ii in axes_info[:self.ndim]])
        self.shape = tuple(reversed(axis_elements))
        self.size = np.product(self.shape)
        self.ndim = len(self.shape)
        return
    
    def __del__(self):
        """VectorOC destructor"""
        if self.remove_file:
            # Removing both header and binary files (using os.system to make module compatible with python3.5)
            os.system("rm -f %s %s" % (self.vecfile, self.binfile))
        return
    
    def getNdArray(self):
        """Function to return Ndarray of the vector"""
        ndarray, _ = sep.read_file(self.vecfile)
        return ndarray
    
    def norm(self, N=2):
        """Function to compute vector N-norm"""
        if N != 2:
            raise NotImplementedError("Norm different than L2 not currently supported")
        # Running Solver_ops to compute norm value
        find = re_dpr.search(RunShellCmd("Solver_ops file1=%s op=dot" % self.vecfile, get_stat=False)[0])
        if find:
            return np.sqrt(float(find.group(1)))
        else:
            raise ValueError("ERROR! Trouble parsing dot product!")
        return
    
    def zero(self):
        """Function to zero out a vector"""
        RunShellCmd("head -c %s </dev/zero > %s" % (self.size * 4, self.binfile),
                    get_stat=False, get_output=False)
        # sys_util.RunShellCmd("Solver_ops file1=%s op=zero"%(self.vecfile),get_stat=False,get_output=False)
        return
    
    def scale(self, sc):
        """Function to scale a vector"""
        RunShellCmd("Solver_ops file1=%s scale1_r=%s op=scale" % (self.vecfile, sc),
                    get_stat=False, get_output=False)
        return
    
    def rand(self, snr=1.0):
        """Fill vector with random number (~U[1,-1]) with a given SNR"""
        # Computing RMS amplitude of the vector
        rms = RunShellCmd("Attr < %s want=rms param=1 maxsize=5000" % (self.vecfile), get_stat=False)[0]
        rms = float(rms.split("=")[1])  # Standard deviation of the signal
        amp_noise = 1.0
        if rms != 0.:
            amp_noise = np.sqrt(3.0 / snr) * rms  # sqrt(3*Power_signal/SNR)
        # Filling file with random number with the proper scale
        RunShellCmd("Noise file=%s rep=1 type=0 var=0.3333333333; Solver_ops file1=%s scale1_r=%s op=scale" % (
            self.vecfile, self.vecfile, amp_noise), get_stat=False, get_output=False)
        return
    
    def clone(self):
        """Function to clone (deep copy) a vector or from a space and creating a copy of the associated header file"""
        # First performing a deep copy of the vector
        vec_clone = deepcopy(self)
        if vec_clone.vecfile is None:
            # Creating header and binary files from vector space
            # Placing temporary file into datapath folder
            tmp_vec = sep.datapath + "clone_tmp_vector" + str(int(time() * 1000000)) + ".H"
            axis_file = ""
            for iaxis, naxis in enumerate(tuple(reversed(vec_clone.shape))):
                axis_file += "n%s=%s " % (iaxis + 1, naxis)
            # Creating temporary vector file
            cmd = "Spike %s | Add scale=0.0 > %s" % (axis_file, tmp_vec)
            RunShellCmd(cmd, get_stat=False, get_output=False)
            vec_clone.vecfile = tmp_vec
            vec_clone.binfile = sep.get_binary(vec_clone.vecfile)
        else:
            # Creating a temporary file with similar name but computer time at the end
            tmp_vec = self.vecfile.split(".H")[0].split("/")[-1]  # Getting filename only
            # Placing temporary file into datapath folder
            tmp_vec = sep.datapath + tmp_vec + "_clone_" + str(int(time() * 1000000)) + ".H"
            tmp_bin = tmp_vec + "@"
            # Copying header and binary files and setting pointers to new file
            copyfile(self.vecfile, tmp_vec)  # Copying header
            copyfile(self.binfile, tmp_bin)  # Copying binary
            vec_clone.vecfile = tmp_vec
            vec_clone.binfile = tmp_bin
            # "Fixing" header file
            with open(vec_clone.vecfile, "a") as fid:
                fid.write("in='%s\n'" % tmp_bin)
        # By default the clone file is going to be removed once the vector is deleted
        vec_clone.remove_file = True
        return vec_clone
    
    def cloneSpace(self):
        """Function to clone vector space only (vector without actual vector binary file by using None values)"""
        vec_space = VectorOC(self.vecfile)
        # Removing header vector file
        vec_space.vecfile = None
        vec_space.binfile = None
        vec_space.remove_file = False
        return vec_space
    
    def checkSame(self, other):
        """Function to check dimensionality of vectors"""
        return self.shape == other.shape
    
    def writeVec(self, filename, mode='w'):
        """Function to write vector to file"""
        # Check writing mode
        if not mode in 'wa':
            raise ValueError("Mode must be appending 'a' or writing 'w' ")
        # writing header/pointer file if not present and not append mode
        if not (os.path.isfile(filename) and mode in 'a'):
            binfile = sep.datapath + filename.split('/')[-1] + '@'
            # Copying SEPlib header file
            copyfile(self.vecfile, filename)
            # Substituting binary file
            with open(filename, 'a') as fid:
                fid.write("\nin='%s'\n" % binfile)
            fid.close()
        else:
            binfile = sep.get_binary(filename)
            if mode in 'a':
                axes = sep.get_axes(filename)
                # Number of vectors already present in the file
                if self.shape == (1,):
                    n_vec = axes[0][0]
                    append_dim = self.ndim
                else:
                    n_vec = axes[self.ndim][0]
                    append_dim = self.ndim + 1
                with open(filename, mode) as fid:
                    fid.write("n%s=%s o%s=0.0 d%s=1.0 \n" % (append_dim, n_vec + 1, append_dim, append_dim))
                fid.close()
        # Writing or Copying binary file
        if not (os.path.isfile(binfile) and mode in 'a'):
            copyfile(self.binfile, binfile)
        else:
            # Writing file if
            with open(binfile, mode + 'b') as fid, open(self.binfile, 'rb') as fid_toread:
                while True:
                    data = fid_toread.read(BUF_SIZE)
                    if not data:
                        break
                    fid.write(data)
            fid.close()
            fid_toread.close()
        return
    
    def copy(self, other):
        """Function to copy vector from input vector"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorOC):
            raise TypeError("ERROR! Provided input vector not a %s!" % self.whoami)
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError(
                "ERROR! Vector dimensionality mismatching: self = %s; other = %s" % (self.shape, other.shape))
        # Copy binary file of input vector
        copyfile(other.binfile, self.binfile)  # Copying binary
        return
    
    def scaleAdd(self, other, sc1=1.0, sc2=1.0):
        """Function to scale a vector"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorOC):
            raise TypeError("ERROR! Provided input vector not a vectorOC!")
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError(
                "ERROR! Vector dimensionality mismatching: self = %s; other = %s" % (self.shape, other.shape))
        # Performing scaling and addition
        cmd = "Solver_ops file1=%s scale1_r=%s file2=%s scale2_r=%s op=scale_addscale" % (
            self.vecfile, sc1, other.vecfile, sc2)
        RunShellCmd(cmd, get_stat=False, get_output=False)
        return
    
    def dot(self, other):
        """Function to compute dot product between two vectors"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorOC):
            raise TypeError("ERROR! Provided input vector not a %s!" % self.whoami)
        # Checking size (must have same number of elements)
        if self.size != other.size:
            raise ValueError("ERROR! Vector size mismatching: self = %s; other = %s" % (self.size, other.size))
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError(
                "ERROR! Vector dimensionality mismatching: self = %s; other = %s" % (self.shape, other.shape))
        # Running Solver_ops to compute norm value
        cmd = "Solver_ops file1=%s file2=%s op=dot" % (self.vecfile, other.vecfile)
        find = re_dpr.search(RunShellCmd(cmd, get_stat=False)[0])
        if find:
            return float(find.group(1))
        else:
            raise ValueError("ERROR! Trouble parsing dot product!")
        return float(out_dot)
    
    def multiply(self, other):
        """Function to multiply element-wise two vectors"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorOC):
            raise TypeError("ERROR! Provided input vector not a %s!" % self.whoami)
        # Checking size (must have same number of elements)
        if self.size != other.size:
            raise ValueError("ERROR! Vector size mismatching: self = %s; other = %s" % (self.size, other.size))
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError(
                "ERROR! Vector dimensionality mismatching: self = %s; other = %s" % (self.shape, other.shape))
        # Performing scaling and addition
        cmd = "Solver_ops file1=%s file2=%s op=multiply" % (self.vecfile, other.vecfile)
        RunShellCmd(cmd, get_stat=False, get_output=False)
        return
    
    def isDifferent(self, other):
        """Function to check if two vectors are identical using M5 hash scheme"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorOC):
            raise TypeError("ERROR! Provided input vector not a %s!" % self.whoami)
        hashmd5_self = hashfile(self.binfile)
        hashmd5_other = hashfile(other.binfile)
        return hashmd5_self != hashmd5_other



import numpy as np
from occamypy import Operator, VectorNumpy, dask

try:
    import pylops
except ImportError:
    import os
    
    os.system("pip install --user pylops")
    import pylops


class FromPylops(Operator):
    
    def __init__(self, domain, range, op):
        """
        Cast a pylops LinearOperator to Operator
        """
        assert isinstance(op, pylops.LinearOperator), "op has to be a pylops.LinearOperator"
        assert op.shape[0] == range.size, "Range and operator rows mismatch"
        assert op.shape[1] == domain.size, "Domain and operator columns mismatch"
        
        self.name = op.__str__()
        self.op = op
        
        super(FromPylops, self).__init__(domain, range)
    
    def __str__(self):
        return self.name.replace('<', '').replace('>', '')
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        x = model.getNdArray().copy().ravel()
        y = self.op.matvec(x)
        data.getNdArray()[:] += y.reshape(data.shape)
        return
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        y = data.getNdArray().copy().ravel()
        x = self.op.rmatvec(y)
        model.getNdArray()[:] += x.reshape(model.shape)
        return


class ToPylops(pylops.LinearOperator):
    
    def __init__(self, op: Operator):
        """
        Cast an in-core Operator to pylops.LinearOperator
        :param op: `occamypy.Operator` object (or child)
        """
        self.explicit = False
        self.shape = (op.range.size, op.domain.size)
        self.dtype = op.domain.getNdArray().dtype
        
        assert isinstance(op, Operator), 'op has to be an Operator'
        self.op = op
        
        # these are just temporary vectors, used by forward and adjoint computations
        self.domain = op.domain.clone()
        self.range = op.range.clone()
    
    def _matvec(self, x):
        x_ = self.domain.clone()
        x_[:] = x.reshape(self.domain.shape).astype(self.dtype)
        y_ = self.range.clone()
        self.op.forward(False, x_, y_)
        return y_.getNdArray().ravel()
    
    def _rmatvec(self, y):
        y_ = self.range.clone()
        y_[:] = y.reshape(self.range.shape).astype(self.dtype)
        x_ = self.domain.clone()
        self.op.adjoint(False, x_, y_)
        return x_.getNdArray().ravel()


class DaskToPylops(pylops.LinearOperator):
    
    def __init__(self, op, chunks):
        """
        Cast a DaskOperator to pylops.LinearOperator
        :param op: `occamypy.DaskOperator` object (or child)
        """
        self.explicit = False
        self.shape = (op.range.size, op.domain.size)
        self.dtype = op.domain.getNdArray()[0].dtype
        
        assert isinstance(op, dask.DaskOperator), 'op has to be a DaskOperator'
        self.op = op
        self.chunks = chunks
    
    def _matvec(self, x):
        # create a dask vector from numpy array
        x_ = dask.DaskVector(self.op.client,
                             VectorNumpy(x.reshape(self.op.domain.shape[0]).astype(self.dtype)),
                             self.chunks)
        y_ = self.op.range.clone()
        self.op.forward(False, x_, y_)
        return y_.getNdArray().ravel()
    
    def _rmatvec(self, y):
        y_ = self.range.clone()
        y_[:] = y.reshape(self.range.shape).astype(self.dtype)
        x_ = self.op.domain.clone()
        self.op.adjoint(False, x_, y_)
        return x_.getNdArray().ravel()


if __name__ == '__main__':
    from occamypy import Scaling
    import matplotlib.pyplot as plt
    
    # shape = (3, 5)
    # x = VectorNumpy(shape).set(1.)
    #
    # # test FromPylops
    # d = np.arange(x.size)
    # D = FromPylops(x, x, pylops.Diagonal(d))
    # D.dotTest(True)
    #
    # # test ToPylops
    # S = ToPylops(Scaling(x, 2.))
    # pylops.utils.dottest(S, x.size, x.size, tol=1e-6, complexflag=0, verb=True)
    # print("\t", np.isclose((S * x.arr.ravel()).reshape(x.shape), 2.*x.arr).all())
    
    # # from interesting pylops
    # # https://pylops.readthedocs.io/en/latest/tutorials/ctscan.html#sphx-glr-tutorials-ctscan-py
    # def radoncurve(x, r, theta):
    #     return (r - ny // 2) / (np.sin(np.deg2rad(theta)) + 1e-15) + np.tan(np.deg2rad(90 - theta)) * x + ny // 2
    #
    # x = np.load('../../tutorials/data/shepp_logan_phantom.npy')
    # x = x / x.max()
    # ny, nx = x.shape
    #
    # ntheta = 150
    # theta = np.linspace(0., 180., ntheta, endpoint=False)
    #
    # RLop = pylops.signalprocessing.Radon2D(np.arange(ny), np.arange(nx),
    #                                        theta, kind=radoncurve,
    #                                        centeredh=True, interp=False,
    #                                        engine='numpy', dtype='float64')
    #
    # y = RLop.H * x.T.ravel()
    # y = y.reshape(ntheta, ny).T
    #
    # xrec = RLop * y.T.ravel()
    # xrec = xrec.reshape(nx, ny).T
    
    # fig, axs = plt.subplots(1, 3, figsize=(10, 4))
    # axs[0].imshow(x, vmin=0, vmax=1, cmap='gray')
    # axs[0].set_title('Model')
    # axs[0].axis('tight')
    # axs[1].imshow(y, cmap='gray')
    # axs[1].set_title('Data')
    # axs[1].axis('tight')
    # axs[2].imshow(xrec, cmap='gray')
    # axs[2].set_title('Adjoint model')
    # axs[2].axis('tight')
    # fig.tight_layout()
    # plt.show()
    
    # # TODO why the fuck they need to be transposed?
    # x_T = VectorNumpy(x).transpose()
    # y_T = VectorNumpy((ntheta, ny))
    # R_T = FromPylops(x_T, y_T, RLop.H)
    # y_T = R_T * x_T
    # xrec_T = R_T.H * y_T
    #
    # fig, axs = plt.subplots(1, 3, figsize=(10, 4))
    # axs[0].imshow(x_T[:], vmin=0, vmax=1, cmap='gray')
    # axs[0].set_title('Model')
    # axs[0].axis('tight')
    # axs[1].imshow(y_T[:], cmap='gray')
    # axs[1].set_title('Data')
    # axs[1].axis('tight')
    # axs[2].imshow(xrec_T[:], cmap='gray')
    # axs[2].set_title('Adjoint model')
    # axs[2].axis('tight')
    # fig.tight_layout()
    # plt.show()
    
    # now with dask
    from occamypy import dask
    
    client = dask.DaskClient(local_params={"processes": True}, n_wrks=4)
    print("Number of workers = %d" % client.getNworkers())
    print("Workers Ids = %s" % client.getWorkerIds())
    
    vec = VectorNumpy(np.load('../../tutorials/data/shepp_logan_phantom.npy') * 1.).scale(1 / 255)
    
    # method 1: instantiate a vector template and spread it using the chunk parameter
    vec_ = dask.DaskVector(client,
                           vector_template=vec,
                           chunks=(2, 3, 5, 10))
    
    # method 2: instantiate multiple vectors and spreading them to the given workers
    vecD = dask.DaskVector(client,
                           vectors=[VectorNumpy(vec[:50]),
                                    VectorNumpy(vec[50:150]),
                                    VectorNumpy(vec[150:])],
                           chunks=(1, 1, 0, 1))
    
    # Instantiating Dask operator: FirstDerivative from pylops
    # D_ = dask.DaskOperator(
    #     client,
    #     FromPylops,
    #     [(v, v, pylops.FirstDerivative(np.prod(shape))) for v, shape in zip(vec_.vecDask, vec_.shape)],
    #     vec_.chunks)
    DD = dask.DaskOperator(
        client,
        FromPylops,
        [(v, v, pylops.FirstDerivative(np.prod(shape))) for v, shape in zip(vecD.vecDask, vecD.shape)],
        vecD.chunks)
    
    # Dot-product test
    DD.dotTest(True)
    # # Power method
    # max_eig = D_.powerMethod()
    # print("\nMaximum eigenvalue = %s" % max_eig)
    
    Spread = dask.DaskSpread(client, vec, [1]*client.getNworkers())
    # Collect a dask vector to a in-core vector
    Collect = dask.DaskCollect(vecD, vec)
    
    # TODO cast a DaskOperator to Pylops
    A_ = dask.DaskOperator(client,
                           Scaling,
                           [(v, 10) for v in vec_.vecDask],
                           vec_.chunks)
    # A_.dotTest(True)
    
    A = DaskToPylops(A_, vec_.chunks)
    y = A * np.ones(vec_.size)
    
    # pylops.utils.dottest(A)
    pylops.utils.dottest(A, *A.shape)
    
    print(0)

__all__ = [
    "ToPylops",
    "FromPylops",
]



from .base import *
from .linear import *
from .nonlinear import *

__all__ = [
    "Operator",
    "Vstack",
    "Hstack",
    "Dstack",
    "Chain",
    "Zero",
    "Identity",
    "Scaling",
    "Diagonal",
    "NonlinearOperator",
    "NonlinearComb",
    "NonlinearSum",
    "NonlinearVstack",
    "VarProOperator",
]



class Logger:
    """System logger class"""
    
    def __init__(self, file, bufsize=1):
        """Initialize writing to a file logfile"""
        self.file = open(file, "a", bufsize)
        return
    
    def __del__(self):
        """Destructor where the log file is closed"""
        self.file.close()
        return
    
    def addToLog(self, msg):
        """Function to write message to log file"""
        self.file.write(msg + "\n")
        return



# Module containing generic Solver and Restart definitions
import atexit
import os
import pickle
import re
import numpy as np
from shutil import rmtree
from copy import deepcopy
import datetime

from occamypy.utils import mkdir, sep
from occamypy import problem as P
from occamypy import VectorSet, VectorOC


class Solver:
    """Solver parent object"""

    # Default class methods/functions
    def __init__(self):
        """Default class constructor for Solver"""
        # Parameter for saving results
        self.save_obj = False
        self.save_res = False
        self.save_grad = False
        self.save_model = False
        self.flush_memory = False

        self.prefix = None

        # Iteration axis-sampling parameters
        self.iter_buffer_size = None
        self.iter_sampling = 1

        # Lists of the results (list and vector Sets)
        self.obj = list()
        self.obj_terms = list()
        self.model = list()
        self.res = list()
        self.grad = list()
        self.modelSet = VectorSet()
        self.resSet = VectorSet()
        self.gradSet = VectorSet()
        self.inv_model = None
        self.iter_written = 0

        # Set Restart object
        self.restart = Restart()
        self.create_msg = False
        # Setting defaults for saving results
        self.setDefaults()
        return

    def __del__(self):
        """Default destructor"""
        return

    def setPrefix(self, prefix):
        """Mutator to change prefix and file names for saving inversion results"""
        self.prefix = prefix
        return

    def setDefaults(self, save_obj=False, save_res=False, save_grad=False, save_model=False, prefix=None,
                    iter_buffer_size=None, iter_sampling=1, flush_memory=False):
        """
        Function to set parameters for result saving.

        :param save_obj         : [False] - boolean; Flag to save objective function values into the list self.obj
        :param save_res         : [False] - boolean; Flag to save residual vectors into the list self.res
        :param save_grad        : [False] - boolean; Flag to save gradient vectors into the list self.grad
        :param save_model       : [False] - boolean; Flag to save model vectors into the list self.model.
                                    It will also say the last inverted model vector into self.inv_model
        :param prefix           : [None] - string; Prefix of the files in which requested results will be saved;
                                    If prefix is None, then nothing is going to be saved on disk
        :param iter_buffer_size : [None] - int; Number of steps to save before flushing results to disk
                                    (by default the solver waits until all iterations are done)
        :param iter_sampling    : [1] - int; Sampling of the iteration axis
        :param flush_memory     : [False] - boolean; Whether to keep results into the object lists or clean those
                                    once inversion is completed or results have been written on disk
        """

        # Parameter for saving results
        self.save_obj = save_obj            # Flag to save objective function value
        self.save_res = save_res            # Flag to save residual vector
        self.save_grad = save_grad          # Flag to save gradient vector
        self.save_model = save_model        # Flag to save model vector
        self.flush_memory = flush_memory    # Keep results in RAM or flush memory every time results are written on disk
        self.overwrite = True               # Flag to overwrite results if first time writing on disk

        # Prefix of the saved files (if provided the results will be written on disk)
        self.prefix = prefix                # Prefix for saving inversion results on disk

        # Iteration axis-sampling parameters
        self.iter_buffer_size = iter_buffer_size  # Number of steps to save before flushing results to disk
        self.iter_sampling = iter_sampling  # Sampling of the iteration axis

        # Lists of the results (list and vector Sets)
        self.obj = np.array([])  # Array for objective function values
        self.obj_terms = np.array([])     # Array for objective function values for each terms
        self.model = list()               # List for model vectors (to save results in-core)
        self.res = list()                 # List for residual vectors (to save results in-core)
        self.grad = list()                # List for gradient vectors (to save results in-core)
        self.modelSet = VectorSet()       # Set for model vectors
        self.resSet = VectorSet()         # Set for residual vectors
        self.gradSet = VectorSet()        # Set for gradient vectors
        self.inv_model = None             # Temporary saved inverted model
        self.overwrite = True             # Flag to overwrite results if first time writing on disk

    def flush_results(self):
        """Flushing internal memory of the saved results"""
        # Lists of the results (list and vector Sets)
        self.obj = np.array([])  # Array for objective function values
        self.obj_terms = np.array([])  # Array for objective function values for each terms
        self.model = list()  # List for model vectors (to save results in-core)
        self.res = list()  # List for residual vectors (to save results in-core)
        self.grad = list()  # List for gradient vectors (to save results in-core)
        self.modelSet = VectorSet()  # Set for model vectors
        self.resSet = VectorSet()  # Set for residual vectors
        self.gradSet = VectorSet()  # Set for gradient vectors
        self.inv_model = None  # Temporary saved inverted model

    def get_restart(self, log_file):
        """
        Function to retrieve restart folder from log file. It enables the user to use restart flag on self.run().
        :param log_file: [None] - string;
        """
        restart_folder = None
        # Obtaining restart folder path
        reg_prog = re.compile("Restart folder: ([^\s]+)")
        if not os.path.isfile(log_file):
            raise OSError("ERROR! No %s file found!" % log_file)
        for line in reversed(open(log_file).readlines()):
            if restart_folder is None:
                find = reg_prog.search(line)
                if find:
                    restart_folder = find.group(1)
        # Setting restart folder if user needs to do so
        if restart_folder is not None:
            self.restart.restart_folder = restart_folder
        else:
            print("WARNING! No restart folder's path was found in %s" % log_file)
        return

    def save_results(self, iiter, problem, **kwargs):
        """
        Method to save results
        :param iiter        : Iteration index
        :param problem      : Problem that is being solved
        :param kwargs       :
        - force_save   : [False]; Flag to ignore iteration sampling
        - force_write  : [False]; Force writing on disk if necessary (used to handle last iteration)
        - model : [problem.model] model to be saved and/or written
        - obj : [problem.obj] objective function to be saved
        """
        if not isinstance(problem, P.Problem):
            raise TypeError("Input variable is not a Problem object")
        force_save = kwargs.get("force_save", False)
        force_write = kwargs.get("force_write", False)
        # Getting a model from arguments if provided (necessary to remove preconditioning)
        mod_save = kwargs.get("model", problem.get_model())
        # Obtaining objective function value
        objf_value = kwargs.get("obj", problem.get_obj(problem.get_model()))
        obj_terms = kwargs.get("obj_terms", problem.obj_terms) if "obj_terms" in dir(problem) else None
        # Save if it is forced to or if the solver hits a sampled iteration number
        # The objective function is saved every iteration if requested
        if self.save_obj:
            self.obj = np.append(self.obj, deepcopy(float(objf_value)))
            # Checking if the objective function has multiple terms
            if obj_terms is not None:
                if len(self.obj_terms) == 0:
                    # First time obj_terms are saved
                    self.obj_terms = np.expand_dims(np.append(self.obj_terms, deepcopy(obj_terms)), axis=0)
                else:
                    self.obj_terms = np.append(self.obj_terms,
                                               np.expand_dims(np.array(deepcopy(obj_terms)), axis=0),
                                               axis=0)
        if iiter % self.iter_sampling == 0 or force_save:
            if self.save_model:
                self.modelSet.append(mod_save)
                # Storing model vector into a temporary vector
                del self.inv_model  # Deallocating previous saved model
                self.inv_model = mod_save.clone()
            if self.save_res:
                res_vec = problem.get_res(problem.get_model())
                self.resSet.append(res_vec)
            if self.save_grad:
                grad = problem.get_grad(problem.get_model())
                self.gradSet.append(grad)
        # Write on disk if necessary or requested
        self._write_steps(force_write)
        return

    def _write_steps(self, force_write=False):
        """Method to write inversion results on disk if forced to or if buffer is filled"""
        # Save results if buffer size is hit
        save = True if force_write or (self.iter_buffer_size is not None and max(len(self.modelSet.vecSet),
                                                                                 len(self.resSet.vecSet),
                                                                                 len(self.gradSet.vecSet))
                                                                             >= self.iter_buffer_size) \
            else False

        # Overwriting results if first time to write on disk
        mode = "w" if self.overwrite else "a"

        if save:
            self.overwrite = False  # Written at least once; do not overwrite files
            # Getting current saved results into an in-core list
            if not self.flush_memory:
                self.model += self.modelSet.vecSet
                self.res += self.resSet.vecSet
                self.grad += self.gradSet.vecSet
            # Writing objective function value on disk if requested
            if self.save_obj and self.prefix is not None:
                obj_file = self.prefix + "_obj.H"  # File name in which the objective function is saved
                sep.write_file(obj_file, self.obj)
                # Writing each term of the objective function
                if len(self.obj_terms) != 0:
                    for iterm in range(self.obj_terms.shape[1]):
                        # File name in which the objective function is saved
                        obj_file = self.prefix + "_obj_comp%s.H" % (iterm + 1)
                        sep.write_file(obj_file, self.obj_terms[:, iterm])
            # Writing current inverted model and model vectors on disk if requested
            if self.save_model and self.prefix is not None:
                inv_mod_file = self.prefix + "_inv_mod.H"  # File name in which the current inverted model is saved
                model_file = self.prefix + "_model.H"  # File name in which the model vector is saved
                self.modelSet.writeSet(model_file, mode=mode)
                self.inv_model.writeVec(inv_mod_file, mode="w") # Writing inverted model file
            # Writing gradient vectors on disk if requested
            if self.save_grad and self.prefix is not None:
                grad_file = self.prefix + "_gradient.H"  # File name in which the gradient vector is saved
                self.gradSet.writeSet(grad_file, mode=mode)
            # Writing residual vectors on disk if requested
            if self.save_res and self.prefix is not None:
                res_file = self.prefix + "_residual.H"  # File name in which the residual vector is saved
                self.resSet.writeSet(res_file, mode=mode)

    def run(self, prblm):
        """Dummy Solver running method"""
        raise NotImplementedError("Implement run Solver in the derived class.")


class Restart:
    """Class for restarting a solver run"""

    def __init__(self):
        """Restart constructor"""
        self.par_dict = dict()
        self.vec_dict = dict()
        # Restart folder in case it is necessary to write restart
        now = datetime.datetime.now()
        restart_folder = sep.datapath + "restart_" + now.isoformat() + "/"
        restart_folder = restart_folder.replace(":", "-")
        self.restart_folder = restart_folder
        # Calling write_restart when python session dies
        atexit.register(self.write_restart)

    def save_vector(self, vec_name, vector_in):
        """Method to save vector for restarting"""
        # Deleting the vector if present in the dictionary
        element = self.vec_dict.pop(vec_name, None)
        if element:
            del element
        self.vec_dict.update({vec_name: vector_in.clone()})

    def retrieve_vector(self, vec_name):
        """Method to retrieve a vector from restart object"""
        return self.vec_dict[vec_name]

    def save_parameter(self, par_name, parameter_in):
        """Method to save parameters for restarting"""
        self.par_dict.update({par_name: parameter_in})
        return

    def retrieve_parameter(self, par_name):
        """Method to retrieve a parameter from restart object"""
        return self.par_dict[par_name]

    def write_restart(self):
        """Restart destructor: it will write vectors on disk if the solver breaks"""
        if bool(self.par_dict) or bool(self.vec_dict):
            # Creating restarting directory
            mkdir(self.restart_folder)
            with open(self.restart_folder + 'restart_obj.pkl', 'wb') as out_file:
                pickle.dump(self, out_file, pickle.HIGHEST_PROTOCOL)
            # Checking if a vectorOC was in the restart and preventing the removal of the vector file
            for vec_name, vec in self.vec_dict.items():
                if isinstance(vec, VectorOC):
                    vec.remove_file = False

    def read_restart(self):
        """Method to read restart object from saved folder"""
        if os.path.isdir(self.restart_folder):
            with open(self.restart_folder + 'restart_obj.pkl', 'rb') as in_file:
                restart = pickle.load(in_file)
            self.par_dict = restart.par_dict
            self.vec_dict = restart.vec_dict
            # Checking if a vectorOC was in the restart and setting the removal of the vector file
            for vec_name, vec in self.vec_dict.items():
                if isinstance(vec, VectorOC):
                    vec.remove_file = True
            # Removing previous restart and deleting read object
            restart.clear_restart()
            del restart

    def clear_restart(self):
        """Method to clear the restart"""
        self.par_dict = dict()
        self.vec_dict = dict()
        # Removing restart folder if existing
        if os.path.isdir(self.restart_folder):
            # Removing folder
            rmtree(self.restart_folder)



import cupy as cp
from occamypy import Operator, Vstack


class FirstDerivative(Operator):
    def __init__(self, model, sampling=1., axis=0, stencil='centered'):
        r"""
        First Derivative with a stencil
            1) 2nd order centered:

            .. math::
                y[i] = 0.5 (x[i+1] - x[i-1]) / dx

            2) 1st order forward:

            .. math::
                y[i] = (x[i+1] - x[i]) / dx

            1) 1st order backward:

            .. math::
                y[i] = 0.5 (x[i] - x[i-1]) / dx

        :param model    : vector class; domain vector
        :param sampling : scalar; sampling step [1.]
        :param axis     : int; axis along which to compute the derivative [0]
        :param stencil  : str; derivative kind (centered, forward, backward)
        """
        self.sampling = sampling
        self.dims = model.getNdArray().shape
        self.axis = axis if axis >= 0 else len(self.dims) + axis
        self.stencil = stencil
        
        if self.stencil == 'centered':
            self.forward = self._forwardC
            self.adjoint = self._adjointC
        elif self.stencil == 'backward':
            self.forward = self._forwardB
            self.adjoint = self._adjointB
        elif self.stencil == 'forward':
            self.forward = self._forwardF
            self.adjoint = self._adjointF
        else:
            raise ValueError("Derivative stencil must be centered, forward or backward")
        
        super(FirstDerivative, self).__init__(model, model)
    
    def __str__(self):
        return "1stDer_%d" % self.axis
    
    def _forwardF(self, add, model, data):
        """Forward operator for the 1st order forward stencil"""
        self.checkDomainRange(model, data)
        if add:
            data_tmp = data.clone()
        data.zero()
        # Getting Ndarrays
        x = model.clone().getNdArray()
        if self.axis > 0:  # need to bring the dim. to derive to first dim
            x = cp.get_array_module(x).swapaxes(x, self.axis, 0)
        y = cp.get_array_module(x).zeros(x.shape)
        
        y[:-1] = (x[1:] - x[:-1]) / self.sampling
        if self.axis > 0:  # reset axis order
            y = cp.get_array_module(x).swapaxes(y, 0, self.axis)
        data.getNdArray()[:] = y
        if add:
            data.scaleAdd(data_tmp)
        return
    
    def _adjointF(self, add, model, data):
        """Adjoint operator for the 1st order forward stencil"""
        self.checkDomainRange(model, data)
        if add:
            model_temp = model.clone()
        model.zero()
        # Getting Ndarrays
        y = data.clone().getNdArray()
        if self.axis > 0:  # need to bring the dim. to derive to first dim
            y = cp.get_array_module(y).swapaxes(y, self.axis, 0)
        x = cp.get_array_module(y).zeros(y.shape)
        
        x[:-1] -= y[:-1] / self.sampling
        x[1:] += y[:-1] / self.sampling
        
        if self.axis > 0:
            x = cp.get_array_module(x).swapaxes(x, 0, self.axis)
        model.getNdArray()[:] = x
        if add:
            model.scaleAdd(model_temp)
        return
    
    def _forwardC(self, add, model, data):
        """Forward operator for the 2nd order centered stencil"""
        self.checkDomainRange(model, data)
        if add:
            data_tmp = data.clone()
        data.zero()
        # Getting Ndarrays
        x = model.clone().getNdArray()
        if self.axis > 0:  # need to bring the dim. to derive to first dim
            x = cp.get_array_module(x).swapaxes(x, self.axis, 0)
        y = cp.get_array_module(x).zeros(x.shape)
        
        y[1:-1] = (.5 * x[2:] - 0.5 * x[:-2]) / self.sampling
        if self.axis > 0:  # reset axis order
            y = cp.get_array_module(x).swapaxes(y, 0, self.axis)
        data.getNdArray()[:] = y
        if add:
            data.scaleAdd(data_tmp)
        return
    
    def _adjointC(self, add, model, data):
        """Adjoint operator for the 2nd order centered stencil"""
        self.checkDomainRange(model, data)
        if add:
            model_temp = model.clone()
        model.zero()
        # Getting Ndarrays
        y = data.clone().getNdArray()
        if self.axis > 0:  # need to bring the dim. to derive to first dim
            y = cp.get_array_module(y).swapaxes(y, self.axis, 0)
        x = cp.get_array_module(y).zeros(y.shape)
        
        x[:-2] -= 0.5 * y[1:-1] / self.sampling
        x[2:] += 0.5 * y[1:-1] / self.sampling
        
        if self.axis > 0:
            x = cp.get_array_module(x).swapaxes(x, 0, self.axis)
        model.getNdArray()[:] = x
        if add:
            model.scaleAdd(model_temp)
        return
    
    def _forwardB(self, add, model, data):
        """Forward operator for the 1st order backward stencil"""
        self.checkDomainRange(model, data)
        if add:
            data_tmp = data.clone()
        data.zero()
        # Getting Ndarrays
        x = model.clone().getNdArray()
        if self.axis > 0:  # need to bring the dim. to derive to first dim
            x = cp.get_array_module(x).swapaxes(x, self.axis, 0)
        y = cp.get_array_module(x).zeros(x.shape)
        
        y[1:] = (x[1:] - x[:-1]) / self.sampling
        if self.axis > 0:  # reset axis order
            y = cp.get_array_module(y).swapaxes(y, 0, self.axis)
        data.getNdArray()[:] = y
        if add:
            data.scaleAdd(data_tmp)
        return
    
    def _adjointB(self, add, model, data):
        """Adjoint operator for the 1st order backward stencil"""
        self.checkDomainRange(model, data)
        if add:
            model_temp = model.clone()
        model.zero()
        # Getting Ndarrays
        y = data.clone().getNdArray()
        if self.axis > 0:  # need to bring the dim. to derive to first dim
            y = cp.get_array_module(y).swapaxes(y, self.axis, 0)
        x = cp.get_array_module(y).zeros(y.shape)
        
        x[:-1] -= y[1:] / self.sampling
        x[1:] += y[1:] / self.sampling
        
        if self.axis > 0:
            x = cp.get_array_module(x).swapaxes(x, 0, self.axis)
        model.getNdArray()[:] = x
        if add:
            model.scaleAdd(model_temp)
        return


class SecondDerivative(Operator):
    def __init__(self, model, sampling=1., axis=0):
        r"""
        Compute 2nd order second derivative

        .. math::
            y[i] = (x[i+1] - 2x[i] + x[i-1]) / dx^2

        :param model    : vector class; domain vector
        :param sampling : scalar; sampling step [1.]
        :param axis     : int; axis along which to compute the derivative [0]
        """
        self.sampling = sampling
        self.data_tmp = model.clone().zero()
        self.dims = model.getNdArray().shape
        self.axis = axis if axis >= 0 else len(self.dims) + axis
        super(SecondDerivative, self).__init__(model, model)
    
    def __str__(self):
        return "2ndDer_%d" % self.axis
    
    def forward(self, add, model, data):
        """Forward operator"""
        self.checkDomainRange(model, data)
        if add:
            self.data_tmp.copy(data)
        data.zero()
        
        # Getting Ndarrays
        x = model.clone().getNdArray()
        if self.axis > 0:  # need to bring the dim. to derive to first dim
            x = cp.get_array_module(x).swapaxes(x, self.axis, 0)
        y = cp.get_array_module(x).zeros_like(x)
        
        y[1:-1] = (x[0:-2] - 2 * x[1:-1] + x[2:]) / self.sampling ** 2
        
        if self.axis > 0:  # reset axis order
            y = cp.get_array_module(y).swapaxes(y, 0, self.axis)
        data.getNdArray()[:] = y
        if add:
            data.scaleAdd(self.data_tmp)
        return
    
    def adjoint(self, add, model, data):
        """Adjoint operator"""
        self.checkDomainRange(model, data)
        if add:
            self.data_tmp.copy(model)
        model.zero()
        
        # Getting numpy arrays
        y = data.clone().getNdArray()
        if self.axis > 0:  # need to bring the dim. to derive to first dim
            y = cp.get_array_module(y).swapaxes(y, self.axis, 0)
        x = cp.get_array_module(y).zeros_like(y)
        
        x[0:-2] += y[1:-1] / self.sampling ** 2
        x[1:-1] -= 2 * y[1:-1] / self.sampling ** 2
        x[2:] += y[1:-1] / self.sampling ** 2
        
        if self.axis > 0:
            x = cp.get_array_module(x).swapaxes(x, 0, self.axis)
        model.getNdArray()[:] = x
        if add:
            model.scaleAdd(self.data_tmp)
        return


class Gradient(Operator):
    def __init__(self, model, sampling=None, stencil=None):
        r"""
        N-Dimensional Gradient operator.

        :param model    : vector class; domain vector
        :param sampling : tuple; sampling step [1]
        :param stencil  : str or list of str; stencil kind for each direction ['centered']
        """
        self.dims = model.getNdArray().shape
        self.sampling = sampling if sampling is not None else tuple([1] * len(self.dims))
        
        if stencil is None:
            self.stencil = tuple(['centered'] * len(self.dims))
        elif isinstance(stencil, str):
            self.stencil = tuple([stencil] * len(self.dims))
        elif isinstance(stencil, tuple) or isinstance(stencil, list):
            self.stencil = stencil
        assert len(self.sampling) == len(self.stencil) != 0, "There is something wrong with the dimensions"
        
        self.op = Vstack([FirstDerivative(model, sampling=self.sampling[d], axis=d)
                          for d in range(len(self.dims))])
        
        super(Gradient, self).__init__(domain=self.op.domain, range=self.op.range)
    
    def __str__(self):
        return "Gradient"
    
    def forward(self, add, model, data):
        return self.op.forward(add, model, data)
    
    def adjoint(self, add, model, data):
        return self.op.adjoint(add, model, data)
    
    def merge_directions(self, add, model, data, iso=True):
        """
        Merge the different directional contributes, using the L2 norm (iso=True) or the simple sum (iso=False)
        """
        self.range.checkSame(model)
        if not add:
            data.zero()
        
        if iso:
            for v in model.vecs:
                data.scaleAdd(v.clone().pow(2), 1., 1.)
                data.pow(.5)
        else:
            for v in model.vecs:
                data.scaleAdd(v, 1., 1.)


class Laplacian(Operator):
    def __init__(self, model, axis=None, weights=None, sampling=None):
        r"""
        Laplacian operator.
        The input parameters are tailored for >2D, but it works also for 1D.

        :param model    : vector class; domain vector
        :param axis     : tuple; axis along which to compute the derivative [all]
        :param weights  : tuple; scalar weights for the axis [1 for each model axis]
        :param sampling : tuple; sampling step [1 for each model axis]
        """
        self.dims = model.getNdArray().shape
        self.axis = axis if axis is not None else tuple(range(len(self.dims)))
        self.sampling = sampling if sampling is not None else tuple([1] * len(self.dims))
        self.weights = weights if weights is not None else tuple([1] * len(self.dims))
        
        assert len(self.axis) == len(self.weights) == len(self.sampling) != 0, \
            "There is something wrong with the dimensions"
        
        self.data_tmp = model.clone().zero()
        
        self.op = self.weights[0] * SecondDerivative(model, sampling=self.sampling[0], axis=self.axis[0])
        for d in range(1, len(self.axis)):
            self.op += self.weights[d] * SecondDerivative(model, sampling=self.sampling[d], axis=self.axis[d])
        super(Laplacian, self).__init__(model, model)
    
    def __str__(self):
        return "Laplace "
    
    def forward(self, add, model, data):
        return self.op.forward(add, model, data)
    
    def adjoint(self, add, model, data):
        return self.op.adjoint(add, model, data)



from math import log, ceil
from itertools import accumulate
import numpy as np
from occamypy import Operator, VectorBaseIC
from .signal import ZeroPad
from scipy.ndimage import gaussian_filter


def slope_estimate(d: np.ndarray, dz: float, dx: float, smooth: float = 20) -> (np.ndarray, np.ndarray):
    r"""Local slope estimation

    Local slopes are estimated using the *Structure Tensor* algorithm [1]_.
    Note that slopes are returned as :math:`arctan(\theta)` where
    :math:`\theta` is an angle defined in a RHS coordinate system with z axis
    pointing upward.

    Parameters
    ----------
    d : Input dataset of size :math:`n_z \times n_x`
    dz : Sampling in z-axis
    dx : Sampling in x-axis
    smooth : Lenght of smoothing filter to be applied to the estimated gradients

    Returns
    -------
    slopes : :obj:`np.ndarray`
        Estimated local slopes
    linearity : :obj:`np.ndarray`
        Estimated linearity

    Notes
    -----
    For each pixel of the input dataset :math:`\mathbf{d}` the local gradients
    :math:`d \mathbf{d} / dz` and :math:`g_z = d\mathbf{d} \ dx` are computed
    and used to define the following three quantities:
    :math:`g_{zz} = (d\mathbf{d} / dz) ^ 2`,
    :math:`g_{xx} = (d\mathbf{d} / dx) ^ 2`, and
    :math:`g_{zx} = d\mathbf{d} / dz * d\mathbf{d} / dx`. Such quantities are
    spatially smoothed and at each pixel their smoothed versions are
    arranged in a :math:`2 \times 2` matrix called the *smoothed
    gradient-square tensor*:

    .. math::
        \mathbf{G} =
        \begin{bmatrix}
           g_{zz}  & g_{zx} \\
           g_{zx}  & g_{xx}
        \end{bmatrix}


    Local slopes can be expressed as
    :math:`p = arctan(\frac{\lambda_{max} - g_{xx}}{g_{zx}})`.

    .. [1] Van Vliet, L. J.,  Verbeek, P. W., "Estimators for orientation and
        anisotropy in digitized images", Journal ASCI Imaging Workshop. 1995.

    """
    nz, nx = d.shape
    gz, gx = np.gradient(d, dz, dx)
    gzz, gzx, gxx = gz * gz, gz * gx, gx * gx
    
    # smoothing
    gzz = gaussian_filter(gzz, sigma=smooth)
    gzx = gaussian_filter(gzx, sigma=smooth)
    gxx = gaussian_filter(gxx, sigma=smooth)
    
    slopes = np.zeros(d.shape)
    linearity = np.zeros(d.shape)
    for iz in range(nz):
        for ix in range(nx):
            l1 = 0.5 * (gzz[iz, ix] + gxx[iz, ix]) + \
                 0.5 * np.sqrt((gzz[iz, ix] - gxx[iz, ix]) ** 2 +
                               4 * gzx[iz, ix] ** 2)
            l2 = 0.5 * (gzz[iz, ix] + gxx[iz, ix]) - \
                 0.5 * np.sqrt((gzz[iz, ix] - gxx[iz, ix]) ** 2 +
                               4 * gzx[iz, ix] ** 2)
            slopes[iz, ix] = np.arctan((l1 - gzz[iz, ix]) / gzx[iz, ix])
            linearity[iz, ix] = 1 - l2 / l1
    return slopes, linearity


def _predict_trace(trace: np.ndarray, t:np.ndarray, dt: float, dx: float, slope:np.ndarray, adj:bool=False) -> np.ndarray:
    r"""Slope-based trace prediction.

    Resample a trace to a new time axis defined by the local slopes along the
    trace. Slopes do implicitly represent a time-varying time delay
    :math:`\Delta t (t) = dx*s(t)`.

    The input trace is interpolated using sinc-interpolation to a new time
    axis given by the following formula: :math:`t_{new} = t + dx*s(t)`.

    Parameters
    ----------
    trace : :obj:`numpy.ndarray`
        Trace
    t : :obj:`numpy.ndarray`
        Time axis
    dt : :obj:`float`
        Time axis sampling
    dx : :obj:`float`
        Spatial axis sampling
    slope : :obj:`numpy.ndarray`
        Slope field
    adj : :obj:`bool`, optional
        Perform forward (``False``) or adjoint (``True``) operation

    Returns
    -------
    tracenew : :obj:`numpy.ndarray`
        Resampled trace

    """
    newt = t - dx * slope
    sinc = np.tile(newt, (len(newt), 1)) - \
           np.tile(t[:, np.newaxis], (1, len(newt)))
    if adj:
        tracenew = np.dot(trace, np.sinc(sinc / dt).T)
    else:
        tracenew = np.dot(trace, np.sinc(sinc / dt))
    return tracenew


def _predict_haar(traces, dt, dx, slopes, repeat=0, backward=False, adj=False):
    """Predict set of traces given time-varying slopes (Haar basis function)

    A set of input traces are resampled based on local slopes. If the number
    of traces in ``slopes`` is twice the number of traces in ``traces``, the
    resampling is done only once per trace. If the number of traces in
    ``slopes`` is a multiple of 2 of the number of traces in ``traces``,
    the prediction is done recursively or in other words the output traces
    are obtained by resampling the input traces followed by ``repeat-1``
    further resampling steps of the intermediate results.

    Parameters
    ----------
    traces : :obj:`numpy.ndarray`
        Input traces of size :math:`n_x \times n_t`
    dt : :obj:`float`
        Time axis sampling of the slope field
    dx : :obj:`float`
        Spatial axis sampling of the slope field
    slopes: :obj:`numpy.ndarray`
        Slope field of size :math:`n_x * 2^{repeat} \times n_t`
    repeat : :obj:`int`, optional
        Number of repeated predictions
    backward : :obj:`bool`, optional
        Predicted trace is on the right (``False``) or on the left (``True``)
        of input trace
    adj : :obj:`bool`, optional
        Perform forward (``False``) or adjoint (``True``) operation

    Returns
    -------
    pred : :obj:`numpy.ndarray`
        Predicted traces

    """
    if backward:
        iback = 1
        idir = -1
    else:
        iback = 0
        idir = 1
    slopejump = 2 ** (repeat + 1)
    repeat = 2 ** repeat

    nx, nt = traces.shape
    t = np.arange(nt) * dt
    pred = np.zeros_like(traces)
    for ix in range(nx):
        pred_tmp = traces[ix]
        if adj:
            for irepeat in range(repeat - 1, -1, -1):
                #print('Slope at', ix * slopejump + iback * repeat + idir * irepeat)
                pred_tmp = \
                    _predict_trace(pred_tmp, t, dt, idir * dx,
                                   slopes[ix * slopejump + iback * repeat + idir * irepeat],
                                   adj=True)
        else:
            for irepeat in range(repeat):
                #print('Slope at', ix * slopejump + iback * repeat + idir * irepeat)
                pred_tmp = \
                    _predict_trace(pred_tmp, t, dt, idir * dx,
                                   slopes[ix * slopejump + iback * repeat + idir * irepeat])
        pred[ix] = pred_tmp
    return pred


def _predict_lin(traces, dt, dx, slopes, repeat=0, backward=False, adj=False):
    """Predict set of traces given time-varying slopes (Linear basis function)

    See _predict_haar for details.
    """
    if backward:
        iback = 1
        idir = -1
    else:
        iback = 0
        idir = 1
    slopejump = 2 ** (repeat + 1)
    repeat = 2 ** repeat

    nx, nt = traces.shape
    t = np.arange(nt) * dt
    pred = np.zeros_like(traces)
    for ix in range(nx):
        pred_tmp = traces[ix]
        #print('Data+ at', ix * slopejump)
        if adj:
            if not ((ix == 0 and not backward) or (ix == nx - 1 and backward)):
                pred_tmp1 = traces[ix - idir]
            #if ix > 0: print('Data- at', (ix - idir) * slopejump)
            for irepeat in range(repeat - 1, -1, -1):
                if (ix == 0 and not backward) or (ix == nx - 1 and backward):
                    #print('Slope+ at', ix * slopejump + iback * repeat + idir * irepeat)
                    pred_tmp = \
                        _predict_trace(pred_tmp, t, dt, idir * dx,
                                       slopes[ix * slopejump + iback * repeat + idir * irepeat],
                                       adj=True)
                    pred_tmp1 = 0
                else:
                    #print('Slope+ at', ix * slopejump + iback * repeat + idir * irepeat)
                    #print('Slope- at', ix * slopejump + iback * repeat - idir * irepeat)
                    pred_tmp = \
                        _predict_trace(pred_tmp, t, dt, idir * dx,
                                       slopes[ix * slopejump + iback * repeat + idir * irepeat],
                                       adj=True)
                    pred_tmp1 = \
                        _predict_trace(pred_tmp1, t, dt, (-idir) * dx,
                                       slopes[ix * slopejump + iback * repeat - idir * irepeat],
                                       adj=True)
        else:
            if not ((ix == nx - 1 and not backward) or (ix == 0 and backward)):
                pred_tmp1 = traces[ix + idir]
            #if ix < nx - 1: print('Data- at', (ix + idir) * slopejump)
            for irepeat in range(repeat):
                if (ix == nx - 1 and not backward) or (ix == 0 and backward):
                    #print('Slope+ at', ix * slopejump + iback * repeat + idir * irepeat)
                    pred_tmp = \
                        _predict_trace(pred_tmp, t, dt, idir * dx,
                                       slopes[ix * slopejump + iback * repeat + idir * irepeat])
                    pred_tmp1 = 0
                else:
                    #print('Slope+ at', ix * slopejump + iback * repeat + idir * irepeat)
                    #print('Slope- at', (ix + idir) * slopejump + iback * repeat - idir * irepeat)
                    pred_tmp = \
                        _predict_trace(pred_tmp, t, dt, idir * dx,
                                       slopes[ix * slopejump + iback * repeat + idir * irepeat])
                    pred_tmp1 = \
                        _predict_trace(pred_tmp1, t, dt, (-idir) * dx,
                                       slopes[(ix + idir) * slopejump + iback * repeat - idir * irepeat])

        #if (adj and ((ix == 0 and not backward) or (ix == nx - 1 and backward))) or
        #    (ix == nx - 1 and not backward) or (ix == 0 and backward):
        #    pred[ix] = pred_tmp
        #else:
        if ix == nx - 1:
            pred[ix] = pred_tmp + pred_tmp1 / 2.
        else:
            pred[ix] = (pred_tmp + pred_tmp1) / 2.
    return pred


class Seislet(Operator):
    r"""2D Seislet operator.

    Apply 2D-Seislet Transform to an input array given an
    estimate of its local ``slopes``. In forward mode, the input array is
    reshaped into a two-dimensional array of size :math:`n_x \times n_t` and
    the transform is performed along the first (spatial) axis (see Notes for
    more details).

    Parameters
    ----------
    slopes : Slope field of size :math:`n_t \times n_x`.
    dt : Sampling steps in t-axis.
    dx : Sampling steps in x-axis.
    level : Number of scaling levels (must be >=0).
    kind : Basis function used for predict and update steps: ``haar`` or
        ``linear``.
    inverse : Apply inverse transform when invoking the adjoint (``True``)
        or not (``False``). Note that in some scenario it may be more
        appropriate to use the exact inverse as adjoint of the Seislet
        operator even if this is not an orthogonal operator and the dot-test
        would not be satisfied (see Notes for details). Otherwise, the user
        can access the inverse directly as method of this class.

    Raises
    ------
    NotImplementedError
        If ``kind`` is different from haar or linear

    Notes
    -----
    The Seislet transform [1]_ is implemented using the lifting scheme.

    In its simplest form (i.e., corresponding to the Haar basis function for
    the Wavelet transform) the input dataset is separated into even
    (:math:`\mathbf{e}`) and odd (:math:`\mathbf{o}`) traces. Even traces are
    used to forward predict the odd traces using local slopes and the
    new odd traces (also referred to as residual) is defined as:

    .. math::
        \mathbf{o}^{i+1} = \mathbf{r}^i = \mathbf{o}^i - P(\mathbf{e}^i)

    where :math:`P = P^+` is the slope-based forward prediction operator
    (which is here implemented as a sinc-based resampling).
    The residual is then updated and summed to the even traces to obtain the
    new even traces (also referred to as coarse representation):

    .. math::
        \mathbf{e}^{i+1} = \mathbf{c}^i = \mathbf{e}^i + U(\mathbf{o}^{i+1})

    where :math:`U = P^- / 2` is the update operator which performs a
    slope-based backward prediction. At this point
    :math:`\mathbf{e}^{i+1}` becomes the new data and the procedure is repeated
    `level` times (at maximum until :math:`\mathbf{e}^{i+1}` is a single trace.
    The Seislet transform is effectively composed of all residuals and
    the coarsest data representation.

    In the inverse transform the two operations are reverted. Starting from the
    coarsest scale data representation :math:`\mathbf{c}` and residual
    :math:`\mathbf{r}`, the even and odd parts of the previous scale are
    reconstructed as:

    .. math::
        \mathbf{e}^i = \mathbf{c}^i - U(\mathbf{r}^i)
        = \mathbf{e}^{i+1} - U(\mathbf{o}^{i+1})

    and:

    .. math::
        \mathbf{o}^i  = \mathbf{r}^i + P(\mathbf{e}^i)
        = \mathbf{o}^{i+1} + P(\mathbf{e}^i)

    A new data is formed by interleaving :math:`\mathbf{e}^i` and
    :math:`\mathbf{o}^i` and the procedure repeated until the new data as the
    same number of traces as the original one.

    Finally the adjoint operator can be easily derived by writing the lifting
    scheme in a matricial form:

    .. math::
        \begin{bmatrix}
           \mathbf{r}_1  \\ \mathbf{r}_2  \\ ... \\ \mathbf{r}_N \\
           \mathbf{c}_1  \\ \mathbf{c}_2  \\ ... \\ \mathbf{c}_N
        \end{bmatrix} =
        \begin{bmatrix}
           \mathbf{I} & \mathbf{0} & ... & \mathbf{0} & -\mathbf{P} & \mathbf{0}  & ... & \mathbf{0}  \\
           \mathbf{0} & \mathbf{I} & ... & \mathbf{0} & \mathbf{0}  & -\mathbf{P} & ... & \mathbf{0}  \\
           ...        & ...        & ... & ...        & ...         & ...         & ... & ...         \\
           \mathbf{0} & \mathbf{0} & ... & \mathbf{I} & \mathbf{0}  & \mathbf{0}  & ... & -\mathbf{P} \\
           \mathbf{U} & \mathbf{0} & ... & \mathbf{0} & \mathbf{I}-\mathbf{UP} & \mathbf{0}  & ... & \mathbf{0}  \\
           \mathbf{0} & \mathbf{U} & ... & \mathbf{0} & \mathbf{0}  & \mathbf{I}-\mathbf{UP} & ... & \mathbf{0}  \\
           ...        & ...        & ... & ...        & ...         & ...         & ... & ...         \\
           \mathbf{0} & \mathbf{0} & ... & \mathbf{U} & \mathbf{0}  & \mathbf{0}  & ... & \mathbf{I}-\mathbf{UP} \\
        \end{bmatrix}
        \begin{bmatrix}
           \mathbf{o}_1  \\ \mathbf{o}_2  \\ ... \\ \mathbf{o}_N \\
           \mathbf{e}_1  \\ \mathbf{e}_2  \\ ... \\ \mathbf{e}_N \\
        \end{bmatrix}

    Transposing the operator leads to:

    .. math::
        \begin{bmatrix}
           \mathbf{o}_1  \\ \mathbf{o}_2  \\ ... \\ \mathbf{o}_N \\
           \mathbf{e}_1  \\ \mathbf{e}_2  \\ ... \\ \mathbf{e}_N \\
        \end{bmatrix} =
        \begin{bmatrix}
           \mathbf{I} & \mathbf{0} & ... & \mathbf{0} & -\mathbf{U^T} & \mathbf{0}  & ... & \mathbf{0}  \\
           \mathbf{0} & \mathbf{I} & ... & \mathbf{0} & \mathbf{0} & -\mathbf{U^T} & ... & \mathbf{0}  \\
           ...        & ...        & ... & ...        & ...        & ...        & ... & ...         \\
           \mathbf{0} & \mathbf{0} & ... & \mathbf{I} & \mathbf{0} & \mathbf{0} & ... & -\mathbf{U^T} \\
           \mathbf{P^T} & \mathbf{0} & ... & \mathbf{0} & \mathbf{I}-\mathbf{P^TU^T} & \mathbf{0} & ... & \mathbf{0}  \\
           \mathbf{0} & \mathbf{P^T} & ... & \mathbf{0} & \mathbf{0} & \mathbf{I}-\mathbf{P^TU^T} & ... & \mathbf{0}  \\
           ...        & ...        & ... & ...          & ...        & ...        & ... & ...         \\
           \mathbf{0} & \mathbf{0} & ... & \mathbf{P^T} & \mathbf{0} & \mathbf{0} & ... & \mathbf{I}-\mathbf{P^TU^T} \\
        \end{bmatrix}
        \begin{bmatrix}
           \mathbf{r}_1  \\ \mathbf{r}_2  \\ ... \\ \mathbf{r}_N \\
           \mathbf{c}_1  \\ \mathbf{c}_2  \\ ... \\ \mathbf{c}_N
        \end{bmatrix}

    which can be written more easily in the following two steps:

    .. math::
        \mathbf{o} = \mathbf{r} + \mathbf{U}^H\mathbf{c}

    and:

    .. math::
        \mathbf{e} = \mathbf{c} - \mathbf{P}^H(\mathbf{r} + \mathbf{U}^H(\mathbf{c})) =
                     \mathbf{c} - \mathbf{P}^H\mathbf{o}

    Similar derivations follow for more complex wavelet bases.

    .. [1] Fomel, S.,  Liu, Y., "Seislet transform and seislet frame",
       Geophysics, 75, no. 3, V25-V38. 2010.

    """
    
    def __init__(self, slopes: VectorBaseIC, dt: float = 1., dx: float = 1.,
                 level: bool = None, kind: str = 'haar', inverse: bool = False):
        
        # define predict and update steps
        if kind == 'haar':
            self.predict_fn = _predict_haar
        elif kind == 'linear':
            self.predict_fn = _predict_lin
        else:
            raise NotImplementedError('kind should be haar or linear')
    
        # define padding for length to be power of 2
        dims = slopes.shape
        ndimpow2 = 2 ** ceil(log(dims[0], 2))
        pad = [(0, 0)] * len(dims)
        pad[0] = (0, ndimpow2 - dims[0])
        self.padding = ZeroPad(slopes, pad)
        self.dims = list(dims)
        self.dims[0] = ndimpow2
        self.nx, self.nt = self.dims
    
        # define levels
        nlevels_max = int(np.log2(self.dims[0]))
        self.levels_size = np.flip(np.array([2 ** i for i in range(nlevels_max)]))
        if level is not None:
            self.levels_size = self.levels_size[:level]
        else:
            self.levels_size = self.levels_size[:-1]
            level = nlevels_max - 1
        self.level = level
        self.levels_cum = np.cumsum(self.levels_size)
        self.levels_cum = np.insert(self.levels_cum, 0, 0)
    
        self.dx, self.dt = dx, dt
        self.slopes = self.padding * slopes
        self.slopes_arr = self.slopes.getNdArray()
        self.inverse = inverse
        super(Seislet, self).__init__(domain=slopes, range=self.slopes)
    
    def forward(self, add, model, data):
        """Forward operator"""
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        # get Ndarrays
        x = (self.padding * model).getNdArray()
        
        # compute with numpy
        y = np.zeros(self.nt, (np.sum(self.levels_size) + self.levels_size[-1], self.nt))
        for ilevel in range(self.level):
            odd = x[:, 1::2]
            even = x[:, ::2]
            res = odd - self.predict_fn(even, dt=self.dt, dx=self.dx, slopes=self.slopes_arr, repeat=ilevel, backward=False)
            x = even + self.predict_fn(res, dt=self.dt, dx=self.dx, slopes=self.slopes_arr, repeat=ilevel, backward=True) / 2.
            y[self.levels_cum[ilevel]:self.levels_cum[ilevel + 1]] = res
        y[self.levels_cum[-1]:] = x
        
        # write to data
        data.getNdArray()[:] += y.T
        return
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        
        x = data.getNdArray()
        
        # now we work with np.ndarrays
        if not self.inverse:
            y = x[self.levels_cum[-1]:]
            for ilevel in range(self.level, 0, -1):
                res = x[self.levels_cum[ilevel - 1]:self.levels_cum[ilevel]]
                odd = res + self.predict_fn(y, self.dt, self.dx, self.slopes_arr,
                                            repeat=ilevel - 1, backward=True,
                                            adj=True) / 2.
                even = y - self.predict_fn(odd, self.dt, self.dx, self.slopes_arr,
                                           repeat=ilevel - 1, backward=False,
                                           adj=True)
                y = np.zeros((2 * even.shape[0], self.nt))
                y[1::2] = odd
                y[::2] = even
        else:
            y = x[self.levels_cum[-1]:]
            for ilevel in range(self.level, 0, -1):
                res = x[self.levels_cum[ilevel - 1]:self.levels_cum[ilevel]]
                even = y - self.predict_fn(res, self.dt, self.dx, self.slopes_arr,
                                           repeat=ilevel - 1, backward=True) / 2.
                odd = res + self.predict_fn(even, self.dt, self.dx, self.slopes_arr,
                                            repeat=ilevel - 1, backward=False)
                y = np.zeros((2 * even.shape[0], self.nt))
                y[1::2] = odd
                y[::2] = even
        
        # convert to VectorIC
        y = self.padding.T * VectorBaseIC(y)
        model + y
        return


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    from pathlib import Path
    
    PLOT = False
    
    inputfile = Path('../../tutorials/data/sigmoid.npz')
    # we want a (nt, nx) vector, C-ordered
    data = VectorBaseIC(np.load(inputfile)['sigmoid'])
    
    nx, nt = data.shape
    dt, dx = 0.004, 8
    t, x = np.arange(nt) * dt, np.arange(nx) * dx
    
    # slope estimation (dz,dx)
    slope = data.clone().zero()
    slope.getNdArray()[:] = -slope_estimate(data.getNdArray().T, dz=dt, dx=dx, smooth=6)[0].T
    
    if PLOT:
        clip = 0.5 * np.max(np.abs(data.arr))
        clip_s = np.max(np.abs(slope.arr))
        opts = dict(aspect='auto', extent=(x[0], x[-1], t[-1], t[0]))
        
        fig, axs = plt.subplots(1, 2, figsize=(10, 4), sharey=True, sharex=True)
        axs[0].imshow(data.getNdArray().T, cmap='gray', vmin=-clip, vmax=clip, **opts)
        axs[0].set(xlabel='Position [m]', ylabel='Time [s]', title='Data')
        axs[0].axis('tight')
        im = axs[1].imshow(slope.getNdArray().T, cmap='jet', vmin=-clip_s, vmax=clip_s, **opts)
        axs[1].set(xlabel='Position [m]', title='Slopes')
        axs[1].axis('tight')
        cax = make_axes_locatable(axs[1]).append_axes('right', size='5%', pad=0.1)
        cb = fig.colorbar(im, cax=cax, orientation='vertical')
        cb.set_label('[m/s]')
        fig.tight_layout()
        plt.show()
    
    # Seislet operator
    Sop = Seislet(slope, dt=dt, dx=dx, kind='haar')

    seis = Sop * data
    # seis = seis.reshape(nx, nt)

    nlevels_max = int(np.log2(nx))
    levels_size = np.flip(np.array([2 ** i for i in range(nlevels_max)]))
    levels_cum = np.cumsum(levels_size)

    plt.figure(figsize=(14, 5))
    plt.imshow(seis.getNdArray(), cmap='gray', vmin=-clip, vmax=clip,
               extent=(1, seis.shape[0], t[-1], t[0]))
    for level in levels_cum:
        plt.axvline(level + 0.5, color='w')
    plt.xlabel('Scale')
    plt.ylabel('Time [s]')
    plt.title('Seislet transform')
    plt.colorbar()
    plt.axis('tight')
    plt.tight_layout(pad=.5)
    plt.show()
    
    # keep the strongest 25% of transform coefficients and reconstruct the data
    perc = 0.25
    seis_strong_idx = np.argsort(-np.abs(seis.ravel()))
    seis_strong = np.abs(seis.ravel())[seis_strong_idx]

    fig, ax = plt.subplots()
    ax.plot(range(1, len(seis_strong) + 1),
            seis_strong / seis_strong[0], label='Seislet')
    ax.set(xlabel="n", ylabel="Coefficient strength [%]",
           title="Transform Coefficients")
    ax.axvline(np.rint(len(seis_strong) * perc),
               color='k', label=f'{100 * perc:.0f}%')
    ax.legend()
    fig.tight_layout()

    seis1 = np.zeros_like(seis.ravel())
    seis_strong_idx = seis_strong_idx[:int(np.rint(len(seis_strong) * perc))]
    seis1[seis_strong_idx] = seis.ravel()[seis_strong_idx]
    d_seis = Sop.inverse(seis1)
    d_seis = d_seis.reshape(nx, nt)
    
    if PLOT:
        opts.update(dict(cmap='gray', vmin=-clip, vmax=clip))
        fig, axs = plt.subplots(1, 3, figsize=(14, 7), sharex=True, sharey=True)
        axs[0].imshow(data.T, **opts)
        axs[0].set(title='Data')
        axs[1].imshow(d_seis.T, **opts)
        axs[1].set(title=f'Rec. from Seislet ({100 * perc:.0f}% of coeffs.)')
        axs[2].imshow((data - d_seis).T, **opts)
        axs[2].set(title='Error from Seislet Rec.')
        for i in range(3):
            axs[i].set(xlabel='Position [m]')
        plt.tight_layout(pad=.5)
        plt.show()



from .vector import *
from .operator import *
from .numpy import *
from .problem import *
from .solver import *
from .dask import *
from .utils import *
from .torch import *

if CUPY_ENABLED:
    from .cupy import *



from copy import deepcopy
from sys import version_info
import numpy as np
from occamypy import Vector, VectorNumpy
try:
    import cupy as cp
    from GPUtil import getGPUs, getFirstAvailable
except:
    pass
    

class VectorCupy(Vector):
    """In-core python vector class based on Cupy"""

    def __init__(self, in_content, device=None):
        """
        VectorCupy constructor
        :param in_content: if a cupy vector, the object is build upon the same vector (i.e., same device).
                       Otherwise, the object is created in the selected device.
        :param device: int - GPU id (None for CPU, -1 for most available memory)
        
        This class stores array with C memory order (i.e., row-wise sorting)
        """
        
        if isinstance(in_content, cp.ndarray) or isinstance(in_content, np.ndarray):
            if cp.isfortran(in_content):
                raise TypeError('Input array not a C contiguous array!')
            self.arr = cp.array(in_content, copy=False)
            self.ax_info = None
        elif isinstance(in_content, tuple):  # Tuple size passed to constructor
            # self.arr = cp.zeros(tuple(reversed(in_content)))
            self.arr = cp.empty(in_content)
            self.ax_info = None
        elif isinstance(in_content, VectorNumpy):
            self.arr = in_content.getNdArray().copy()
            self.ax_info = in_content.ax_info
        else:  # Not supported type
            raise ValueError("ERROR! Input variable not currently supported!")
        
        self.setDevice(device)
        
        super(VectorCupy, self).__init__()
        self.shape = self.arr.shape   # Number of elements per axis (tuple)
        self.ndim = self.arr.ndim    # Number of axes (integer)
        self.size = self.arr.size     # Total number of elements (integer)

    def _check_same_device(self, other):
        assert isinstance(self, VectorCupy)
        assert isinstance(other, VectorCupy)
        answer = self.device == other.device
        if not answer:
            raise Warning('The two vectors live in different devices: %s - %s' % (self.device, other.device))
        return answer
    
    @property
    def device(self):
        try:
            return self.arr.device
        except AttributeError:
            return None

    def setDevice(self, devID=0):
        if devID is not None:  # Move to GPU
            if devID == -1:
                devID = getFirstAvailable(order='memory')[0]
            try:
                with cp.cuda.Device(devID):
                    self.arr = cp.asarray(self.arr)
            except AttributeError:
                self.arr.device = None
        else:  # move to CPU
            if self.device is None:  # already on CPU
                pass
            else:
                self.arr = cp.asnumpy(self.arr)
    
    def printDevice(self):
        if self.device is None:
            print('CPU')
        else:
            name = getGPUs()[self.device.id].name
            print('GPU %d - %s' % (self.device.id, name))

    def getNdArray(self):
        """Function to return Ndarray of the vector"""
        return self.arr
    
    def norm(self, N=2):
        """Function to compute vector N-norm using Numpy"""
        return cp.linalg.norm(self.getNdArray().ravel(), ord=N)

    def zero(self):
        """Function to zero out a vector"""
        self.getNdArray().fill(0)
        return self

    def max(self):
        """Function to obtain maximum value in the vector"""
        return self.getNdArray().max()

    def min(self):
        """Function to obtain minimum value in the vector"""
        return self.getNdArray().min()

    def set(self, val):
        """Function to set all values in the vector"""
        self.getNdArray().fill(val)
        return self

    def scale(self, sc):
        """Function to scale a vector"""
        self.getNdArray()[:] *= sc
        return self

    def addbias(self, bias):
        self.getNdArray()[:] += bias
        return self

    def rand(self, snr=1.):
        """Fill vector with random number (~U[1,-1]) with a given SNR"""
        rms = cp.sqrt(cp.mean(cp.square(self.getNdArray())))
        amp_noise = 1.0
        if rms != 0.:
            amp_noise = cp.sqrt(3. / snr) * rms  # sqrt(3*Power_signal/SNR)
        self.getNdArray()[:] = amp_noise * (2. * cp.random.random(self.getNdArray().shape) - 1.)
        return self

    def clone(self):
        """Function to clone (deep copy) a vector from a vector or a Space"""
        vec_clone = deepcopy(self)  # Deep clone of vector
        # Checking if a vector space was provided
        if vec_clone.getNdArray().size == 0:
            vec_clone.arr = cp.zeros(vec_clone.shape, dtype=self.getNdArray().dtype)
        return vec_clone

    def cloneSpace(self):
        """Function to clone vector space only (vector without actual vector array by using empty array of size 0)"""
        arr = cp.empty(0, dtype=self.getNdArray().dtype)
        vec_space = VectorCupy(arr)
        # Cloning space of input vector
        vec_space.shape = self.shape
        vec_space.ndim = self.ndim
        vec_space.size = self.size
        return vec_space

    def checkSame(self, other):
        """Function to check dimensionality of vectors"""
        return self.shape == other.shape
    
    def abs(self):
        self.getNdArray()[:] = cp.abs(self.getNdArray())
        return self

    def sign(self):
        self.getNdArray()[:] = cp.sign(self.getNdArray())
        return self

    def reciprocal(self):
        self.getNdArray()[:] = 1. / self.getNdArray()
        return self

    def maximum(self, other):
        if cp.isscalar(other):
            self.getNdArray()[:] = cp.maximum(self.getNdArray(), other)
            return self
        elif isinstance(other, VectorCupy):
            if not self.checkSame(other):
                raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
            if not self._check_same_device(other):
                raise ValueError('Provided input has to live in the same device')
            self.getNdArray()[:] = cp.maximum(self.getNdArray(), other.getNdArray())
            return self
        else:
            raise TypeError("Provided input has to be either a scalar or a %s!" % self.whoami)

    def conj(self):
        self.getNdArray()[:] = cp.conj(self.getNdArray())
        return self

    def pow(self, power):
        """Compute element-wise power of the vector"""
        self.getNdArray()[:] = self.getNdArray() ** power
        return self

    def real(self):
        """Return the real part of the vector"""
        self.getNdArray()[:] = self.getNdArray().real
        return self

    def imag(self,):
        """Return the imaginary part of the vector"""
        self.getNdArray()[:] = self.getNdArray().imag
        return self

    def copy(self, other):
        """Function to copy vector from input vector"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorCupy):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError("Dimensionality not equal: self = %s; other = %s" % (self.shape, other.shape))
        # Element-wise copy of the input array
        self.getNdArray()[:] = other.getNdArray()
        return self

    def scaleAdd(self, other, sc1=1.0, sc2=1.0):
        """Function to scale a vector"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorCupy):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError("Dimensionality not equal: self = %s; other = %s" % (self.shape, other.shape))
        # Performing scaling and addition
        if not self._check_same_device(other):
            raise ValueError('Provided input has to live in the same device')
        self.getNdArray()[:] = sc1 * self.getNdArray() + sc2 * other.getNdArray()
        return self

    def dot(self, other):
        """Function to compute dot product between two vectors"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorCupy):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking size (must have same number of elements)
        if self.size != other.size:
            raise ValueError("Vector size mismatching: self = %d; other = %d" % (self.size, other.size))
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError("Dimensionality not equal: self = %s; other = %s" % (self.shape, other.shape))
        if not self._check_same_device(other):
            raise ValueError('Provided input has to live in the same device')
        return cp.vdot(self.getNdArray().flatten(), other.getNdArray().flatten())

    def multiply(self, other):
        """Function to multiply element-wise two vectors"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorCupy):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        # Checking size (must have same number of elements)
        if self.size != other.size:
            raise ValueError("Vector size mismatching: self = %d; other = %d" % (self.size, other.size))
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError("Dimensionality not equal: self = %s; other = %s" % (self.shape, other.shape))
        # Performing element-wise multiplication
        if not self._check_same_device(other):
            raise ValueError('Provided input has to live in the same device')
        self.getNdArray()[:] = cp.multiply(self.getNdArray(), other.getNdArray())
        return self

    def isDifferent(self, other):
        """Function to check if two vectors are identical using built-in hash function"""
        # Checking whether the input is a vector or not
        if not isinstance(other, VectorCupy):
            raise TypeError("Provided input vector not a %s!" % self.whoami)
        if not self._check_same_device(other):
            raise ValueError('Provided input has to live in the same device')
        # Using Hash table for python2 and numpy built-in function array_equal otherwise
        if version_info[0] == 2:
            # First make both array buffers read-only
            self.arr.flags.writeable = False
            other.arr.flags.writeable = False
            chcksum1 = hash(self.getNdArray().data)
            chcksum2 = hash(other.getNdArray().data)
            # Remake array buffers writable
            self.arr.flags.writeable = True
            other.arr.flags.writeable = True
            isDiff = (chcksum1 != chcksum2)
        else:
            isDiff = (not cp.equal(self.getNdArray(), other.getNdArray()).all())
        return isDiff

    def clipVector(self, low, high):
        """Function to bound vector values based on input vectors low and high"""
        if not isinstance(low, VectorCupy):
            raise TypeError("Provided input low vector not a %s!" % self.whoami)
        if not isinstance(high, VectorCupy):
            raise TypeError("Provided input high vector not a %s!" % self.whoami)
        self.getNdArray()[:] = cp.minimum(cp.maximum(low.getNdArray(), self.getNdArray()), high.getNdArray())
        return self
    
    def plot(self):
        return self.getNdArray().get()


if __name__ == '__main__':
    from occamypy.cupy.operator import ConvND
    x = VectorCupy((1000, 20000)).set(1.)
    x.printDevice()

    # D = pyCuOperator.FirstDerivative(x)
    # n = x.clone().rand()
    # y = x.clone().set(10) + 0.01 * n
    # S = pyCuOperator.scalingOp(x, 10)
    # xinv = S / y
    # print('Error norm = %.2e' % (xinv.norm() - x.norm()))

    # Test Convolution
    nh = [5, 10]
    hz = cp.exp(-0.1 * cp.linspace(-(nh[0] // 2), nh[0] // 2, nh[0]) ** 2)
    hx = cp.exp(-0.03 * cp.linspace(-(nh[1] // 2), nh[1] // 2, nh[1]) ** 2)
    hz /= cp.trapz(hz)  # normalize the integral to 1
    hx /= cp.trapz(hx)  # normalize the integral to 1
    kernel = hz[:, cp.newaxis] * hx[cp.newaxis, :]
    C = ConvND(x, kernel)
    C.dotTest(True)
    #
    # x = vectorCupy(cp.arange(9).reshape((3, 3)))
    # pad = ((2, 2), (3, 3))
    # P = pyCuOperator.ZeroPad(x, pad)

    print(0)
